from __future__ import absolute_import, division, print_function

try:
    import h5py  # if we import h5py after tables we segfault
except ImportError:
    pass

from pandas import DataFrame
from odo import odo, convert, append, drop, resource
from odo.backends.csv import CSV
from odo.backends.json import JSON, JSONLines

from multipledispatch import halt_ordering, restart_ordering

halt_ordering()  # Turn off multipledispatch ordering

from datashape import dshape, discover
from .utils import ignoring
import warnings
from .expr import (
    Symbol,
    broadcast_collect,
    by,
    cast,
    coalesce,
    count,
    count_values,
    date,
    datetime,
    day,
    distinct,
    distinct, head,
    head,
    hour,
    join,
    label,
    like,
    mean,
    merge,
    microsecond,
    millisecond,
    month,
    ndim,
    nunique,
    relabel,
    sample,
    second,
    selection,
    shape,
    sort,
    summary,
    symbol,
    time,
    transform,
    var,
    year,
)
from .expr.arrays import (tensordot, transpose)
from .expr.functions import *
from .index import create_index
from .interactive import *
from .compute.pmap import set_default_pmap
from .compute.csv import *
from .compute.json import *
from .compute.python import *
from .compute.pandas import *
from .compute.numpy import *
from .compute.core import *
from .compute.core import compute
from .cached import CachedDataset

with ignoring(ImportError):
    from .server import *
with ignoring(ImportError):
    from .sql import *
    from .compute.sql import *
with ignoring(ImportError):
    from .compute.dask import *
with ignoring(ImportError, AttributeError):
    from .compute.spark import *
with ignoring(ImportError, TypeError):
    from .compute.sparksql import *
with ignoring(ImportError):
    from .compute.h5py import *
with ignoring(ImportError):
    from .compute.hdfstore import *
with ignoring(ImportError):
    from .compute.pytables import *
with ignoring(ImportError):
    from .compute.chunks import *
with ignoring(ImportError):
    from .compute.bcolz import *
with ignoring(ImportError):
    from .mongo import *
    from .compute.mongo import *
with ignoring(ImportError):
    from .pytables import *
    from .compute.pytables import *


from .expr import concat  # Some module re-export toolz.concat and * catches it.

restart_ordering()  # Restart multipledispatch ordering and do ordering

inf = float('inf')
nan = float('nan')

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



from __future__ import absolute_import, division, print_function

import numbers
import numpy as np
from functools import partial
from itertools import chain

import datashape
from datashape import (DataShape, Option, Record, Unit, dshape, var, Fixed,
                       Var, promote, object_)
from datashape.predicates import isscalar, iscollection, isrecord
from toolz import (isdistinct, frequencies, concat as tconcat, unique, get,
                   first, compose, keymap)
import toolz.curried.operator as op
from odo.utils import copydoc

from .core import common_subexpression
from .expressions import Expr, ElemWise, label, Field
from .expressions import dshape_method_list
from ..compatibility import zip_longest, _strtypes
from ..utils import listpack


__all__ = ['Concat',
           'concat',
           'Distinct',
           'distinct',
           'Head',
           'head',
           'IsIn',
           'isin',
           'Join',
           'join',
           'Merge',
           'merge',
           'Sample',
           'sample',
           'Shift',
           'shift',
           'Sort',
           'sort',
           'Tail',
           'tail',
           'transform']


class Sort(Expr):

    """ Table in sorted order

    Examples
    --------
    >>> from blaze import symbol
    >>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
    >>> accounts.sort('amount', ascending=False).schema
    dshape("{name: string, amount: int32}")

    Some backends support sorting by arbitrary rowwise tables, e.g.

    >>> accounts.sort(-accounts.amount) # doctest: +SKIP
    """
    _arguments = '_child', '_key', 'ascending'

    def _dshape(self):
        return self._child.dshape

    @property
    def key(self):
        if self._key is () or self._key is None:
            return self._child.fields[0]
        if isinstance(self._key, tuple):
            return list(self._key)
        else:
            return self._key

    def _len(self):
        return self._child._len()

    @property
    def _name(self):
        return self._child._name

    def __str__(self):
        return "%s.sort(%s, ascending=%s)" % (self._child, repr(self._key),
                                              self.ascending)


def sort(child, key=None, ascending=True):
    """ Sort a collection

    Parameters
    ----------
    key : str, list of str, or Expr
        Defines by what you want to sort.

          * A single column string: ``t.sort('amount')``
          * A list of column strings: ``t.sort(['name', 'amount'])``
          * An expression: ``t.sort(-t.amount)``

        If sorting a columnar dataset, the ``key`` is ignored, as it is not
        necessary:

          * ``t.amount.sort()``
          * ``t.amount.sort('amount')``
          * ``t.amount.sort('foobar')``

       are all equivalent.

    ascending : bool, optional
        Determines order of the sort
    """
    if ascending not in (True, False):
        # NOTE: this test is to guard against users saying `x.sort('a', 'b')`
        # when they should have said `x.sort(['a', 'b'])`.
        msg = "ascending must be True or False, given {}"
        raise ValueError(msg.format(ascending))
    if not isrecord(child.dshape.measure):
        if key is None or isinstance(key, _strtypes):
            # Handle this case separately.
            return Sort(child, None, ascending)
        msg = "sort key {!r} not valid for schema {!r}"
        raise ValueError(msg.format(key, child.dshape.measure))
    if key is None and isrecord(child.dshape.measure):
        key = child.dshape.measure.names
    if isinstance(key, (list, tuple)):
        key = keys_to_validate = tuple(key)
    else:
        keys_to_validate = (key,)
    for k in keys_to_validate:
        if k is None:
            msg = "sort key {!r} not valid for schema {!r}"
            raise ValueError(msg.format(k, child.dshape.measure))
        elif isinstance(k, _strtypes):
            if k not in child.dshape.measure.names:
                msg = "sort key {} is not a column of schema {}"
                raise ValueError(msg.format(k, child.dshape.measure))
        elif not isinstance(k, Expr):
            msg = "sort key {} is not a string column name or an expression."
            raise ValueError(msg.format(k))
    return Sort(child, key, ascending)


class Distinct(Expr):

    """ Remove duplicate elements from an expression

    Parameters
    ----------
    on : tuple of :class:`~blaze.expr.expressions.Field`
        The subset of fields or names of fields to be distinct on.

    Examples
    --------
    >>> from blaze import symbol
    >>> t = symbol('t', 'var * {name: string, amount: int, id: int}')
    >>> e = distinct(t)

    >>> data = [('Alice', 100, 1),
    ...         ('Bob', 200, 2),
    ...         ('Alice', 100, 1)]

    >>> from blaze.compute.python import compute
    >>> sorted(compute(e, data))
    [('Alice', 100, 1), ('Bob', 200, 2)]

    Use a subset by passing `on`:

    >>> import pandas as pd
    >>> e = distinct(t, 'name')
    >>> data = pd.DataFrame([['Alice', 100, 1],
    ...                      ['Alice', 200, 2],
    ...                      ['Bob', 100, 1],
    ...                      ['Bob', 200, 2]],
    ...                     columns=['name', 'amount', 'id'])
    >>> compute(e, data)
        name  amount  id
    0  Alice     100   1
    1    Bob     100   1


    """
    _arguments = '_child', 'on'

    def _dshape(self):
        return datashape.var * self._child.dshape.measure

    @property
    def fields(self):
        return self._child.fields

    @property
    def _name(self):
        return self._child._name

    def __str__(self):
        return 'distinct({child}{on})'.format(
            child=self._child,
            on=(', ' if self.on else '') + ', '.join(map(str, self.on))
        )


@copydoc(Distinct)
def distinct(expr, *on):
    fields = frozenset(expr.fields)
    _on = []
    append = _on.append
    for n in on:
        if isinstance(n, Field):
            if n._child.isidentical(expr):
                n = n._name
            else:
                raise ValueError('{0} is not a field of {1}'.format(n, expr))
        if not isinstance(n, _strtypes):
            raise TypeError('on must be a name or field, not: {0}'.format(n))
        elif n not in fields:
            raise ValueError('{0} is not a field of {1}'.format(n, expr))
        append(n)

    return Distinct(expr, tuple(_on))


class _HeadOrTail(Expr):
    _arguments = '_child', 'n'

    def _dshape(self):
        return self.n * self._child.dshape.subshape[0]

    def _len(self):
        return min(self._child._len(), self.n)

    @property
    def _name(self):
        return self._child._name

    def __str__(self):
        return '%s.%s(%d)' % (self._child, type(self).__name__.lower(), self.n)


class Head(_HeadOrTail):

    """ First `n` elements of collection

    Examples
    --------
    >>> from blaze import symbol
    >>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
    >>> accounts.head(5).dshape
    dshape("5 * {name: string, amount: int32}")

    See Also
    --------

    blaze.expr.collections.Tail
    """
    pass


@copydoc(Head)
def head(child, n=10):
    return Head(child, n)


class Tail(_HeadOrTail):
    """ Last `n` elements of collection

    Examples
    --------
    >>> from blaze import symbol
    >>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
    >>> accounts.tail(5).dshape
    dshape("5 * {name: string, amount: int32}")

    See Also
    --------

    blaze.expr.collections.Head
    """
    pass


@copydoc(Tail)
def tail(child, n=10):
    return Tail(child, n)


class Sample(Expr):
    """Random row-wise sample.  Can specify `n` or `frac` for an absolute or
    fractional number of rows, respectively.

    Examples
    --------
    >>> from blaze import symbol
    >>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
    >>> accounts.sample(n=2).dshape
    dshape("var * {name: string, amount: int32}")
    >>> accounts.sample(frac=0.1).dshape
    dshape("var * {name: string, amount: int32}")
    """
    _arguments = '_child', 'n', 'frac'

    def _dshape(self):
        return self._child.dshape

    def __str__(self):
        arg = 'n={}'.format(self.n) if self.n is not None else 'frac={}'.format(self.frac)
        return '%s.sample(%s)' % (self._child, arg)


@copydoc(Sample)
def sample(child, n=None, frac=None):
    if n is frac is None:
        raise TypeError("sample() missing 1 required argument, 'n' or 'frac'.")
    if n is not None and frac is not None:
        raise ValueError("n ({}) and frac ({}) cannot both be specified.".format(n, frac))
    if n is not None:
        n = op.index(n)
        if n < 1:
            raise ValueError("n must be positive, given {}".format(n))
    if frac is not None:
        frac = float(frac)
        if not 0.0 <= frac <= 1.0:
            raise ValueError("sample requires 0 <= frac <= 1.0, given {}".format(frac))
    return Sample(child, n, frac)


def transform(t, replace=True, **kwargs):
    """ Add named columns to table

    >>> from blaze import symbol
    >>> t = symbol('t', 'var * {x: int, y: int}')
    >>> transform(t, z=t.x + t.y).fields
    ['x', 'y', 'z']
    """
    if replace and set(t.fields).intersection(set(kwargs)):
        t = t[[c for c in t.fields if c not in kwargs]]

    args = [t] + [v.label(k) for k, v in sorted(kwargs.items(), key=first)]
    return merge(*args)


def schema_concat(exprs):
    """ Concatenate schemas together.  Supporting both Records and Units

    In the case of Units, the name is taken from expr.name
    """
    new_fields = []
    for c in exprs:
        schema = c.schema[0]
        if isinstance(schema, Record):
            new_fields.extend(schema.fields)
        elif isinstance(schema, (Unit, Option)):
            new_fields.append((c._name, schema))
        else:
            raise TypeError("All schemas must have Record or Unit shape."
                            "\nGot %s" % schema)
    return dshape(Record(new_fields))


class Merge(ElemWise):

    """ Merge many fields together

    Examples
    --------
    >>> from blaze import symbol, label
    >>> accounts = symbol('accounts', 'var * {name: string, x: int, y: real}')
    >>> merge(accounts.name, z=accounts.x + accounts.y).fields
    ['name', 'z']

    To control the ordering of the fields, use ``label``:

    >>> merge(label(accounts.name, 'NAME'), label(accounts.x, 'X')).dshape
    dshape("var * {NAME: string, X: int32}")
    >>> merge(label(accounts.x, 'X'), label(accounts.name, 'NAME')).dshape
    dshape("var * {X: int32, NAME: string}")
    """
    _arguments = '_child', 'children'

    def _schema(self):
        return schema_concat(self.children)

    @property
    def fields(self):
        return list(tconcat(child.fields for child in self.children))

    def _subterms(self):
        yield self
        for i in self.children:
            for node in i._subterms():
                yield node

    def _get_field(self, key):
        for child in self.children:
            if key in child.fields:
                if isscalar(child.dshape.measure):
                    return child
                else:
                    return child[key]

    def _project(self, key):
        if not isinstance(key, (tuple, list)):
            raise TypeError("Expected tuple or list, got %s" % key)
        return merge(*[self[c] for c in key])

    def _leaves(self):
        return list(unique(tconcat(i._leaves() for i in self.children)))


@copydoc(Merge)
def merge(*exprs, **kwargs):
    if len(exprs) + len(kwargs) == 1:
        if exprs:
            return exprs[0]
        if kwargs:
            [(k, v)] = kwargs.items()
            return v.label(k)
    # Get common sub expression
    exprs += tuple(label(v, k) for k, v in sorted(kwargs.items(), key=first))
    child = common_subexpression(*exprs)
    result = Merge(child, exprs)

    if not isdistinct(result.fields):
        raise ValueError(
            "Repeated columns found: " + ', '.join(
                k for k, v in frequencies(result.fields).items() if v > 1
            ),
        )

    return result


def unpack(l):
    """ Unpack items from collections of nelements 1

    >>> unpack('hello')
    'hello'
    >>> unpack(['hello'])
    'hello'
    """
    if isinstance(l, (tuple, list, set)) and len(l) == 1:
        return next(iter(l))
    else:
        return l


class Join(Expr):

    """ Join two tables on common columns

    Parameters
    ----------
    lhs, rhs : Expr
        Expressions to join
    on_left : str, optional
        The fields from the left side to join on.
        If no ``on_right`` is passed, then these are the fields for both
        sides.
    on_right : str, optional
        The fields from the right side to join on.
    how : {'inner', 'outer', 'left', 'right'}
        What type of join to perform.
    suffixes: pair of str
        The suffixes to be applied to the left and right sides
        in order to resolve duplicate field names.

    Examples
    --------
    >>> from blaze import symbol
    >>> names = symbol('names', 'var * {name: string, id: int}')
    >>> amounts = symbol('amounts', 'var * {amount: int, id: int}')

    Join tables based on shared column name

    >>> joined = join(names, amounts, 'id')

    Join based on different column names

    >>> amounts = symbol('amounts', 'var * {amount: int, acctNumber: int}')
    >>> joined = join(names, amounts, 'id', 'acctNumber')

    See Also
    --------

    blaze.expr.collections.Merge
    """
    _arguments = 'lhs', 'rhs', '_on_left', '_on_right', 'how', 'suffixes'
    __inputs__ = 'lhs', 'rhs'

    @property
    def on_left(self):
        on_left = self._on_left
        if isinstance(on_left, tuple):
            return list(on_left)
        return on_left

    @property
    def on_right(self):
        on_right = self._on_right
        if isinstance(on_right, tuple):
            return list(on_right)
        return on_right

    def _schema(self):
        """

        Examples
        --------
        >>> from blaze import symbol
        >>> t = symbol('t', 'var * {name: string, amount: int}')
        >>> s = symbol('t', 'var * {name: string, id: int}')

        >>> join(t, s).schema
        dshape("{name: string, amount: int32, id: int32}")

        >>> join(t, s, how='left').schema
        dshape("{name: string, amount: int32, id: ?int32}")

        Overlapping but non-joined fields append _left, _right

        >>> a = symbol('a', 'var * {x: int, y: int}')
        >>> b = symbol('b', 'var * {x: int, y: int}')
        >>> join(a, b, 'x').fields
        ['x', 'y_left', 'y_right']
        """

        option = lambda dt: dt if isinstance(dt, Option) else Option(dt)

        on_left = self.on_left
        if not isinstance(on_left, list):
            on_left = on_left,

        on_right = self.on_right
        if not isinstance(on_right, list):
            on_right = on_right,

        right_types = keymap(
            dict(zip(on_right, on_left)).get,
            self.rhs.dshape.measure.dict,
        )
        joined = (
            (name, promote(dt, right_types[name], promote_option=False))
            for n, (name, dt) in enumerate(filter(
                compose(op.contains(on_left), first),
                self.lhs.dshape.measure.fields,
            ))
        )

        left = [
            (name, dt) for name, dt in zip(
                self.lhs.fields,
                types_of_fields(self.lhs.fields, self.lhs)
            ) if name not in on_left
        ]

        right = [
            (name, dt) for name, dt in zip(
                self.rhs.fields,
                types_of_fields(self.rhs.fields, self.rhs)
            ) if name not in on_right
        ]

        # Handle overlapping but non-joined case, e.g.
        left_other = set(name for name, dt in left if name not in on_left)
        right_other = set(name for name, dt in right if name not in on_right)
        overlap = left_other & right_other

        left_suffix, right_suffix = self.suffixes
        left = ((name + left_suffix if name in overlap else name, dt)
                for name, dt in left)
        right = ((name + right_suffix if name in overlap else name, dt)
                 for name, dt in right)

        if self.how in ('right', 'outer'):
            left = ((name, option(dt)) for name, dt in left)
        if self.how in ('left', 'outer'):
            right = ((name, option(dt)) for name, dt in right)

        return dshape(Record(chain(joined, left, right)))

    def _dshape(self):
        # TODO: think if this can be generalized
        return var * self.schema


def types_of_fields(fields, expr):
    """ Get the types of fields in an expression

    Examples
    --------
    >>> from blaze import symbol
    >>> expr = symbol('e', 'var * {x: int64, y: float32}')
    >>> types_of_fields('y', expr)
    ctype("float32")

    >>> types_of_fields(['y', 'x'], expr)
    (ctype("float32"), ctype("int64"))

    >>> types_of_fields('x', expr.x)
    ctype("int64")
    """
    if isinstance(expr.dshape.measure, Record):
        return get(fields, expr.dshape.measure)
    else:
        if isinstance(fields, (tuple, list, set)):
            assert len(fields) == 1
            fields, = fields
        assert fields == expr._name
        return expr.dshape.measure


@copydoc(Join)
def join(lhs, rhs, on_left=None, on_right=None,
         how='inner', suffixes=('_left', '_right')):
    if not on_left and not on_right:
        on_left = on_right = unpack(list(sorted(
            set(lhs.fields) & set(rhs.fields),
            key=lhs.fields.index)))
    if not on_right:
        on_right = on_left
    if isinstance(on_left, tuple):
        on_left = list(on_left)
    if isinstance(on_right, tuple):
        on_right = list(on_right)
    if not on_left or not on_right:
        raise ValueError(
            "Can not Join.  No shared columns between %s and %s" % (lhs, rhs),
        )
    left_types = listpack(types_of_fields(on_left, lhs))
    right_types = listpack(types_of_fields(on_right, rhs))
    if len(left_types) != len(right_types):
        raise ValueError(
            'Length of on_left=%d not equal to length of on_right=%d' % (
                len(left_types), len(right_types),
            ),
        )

    for n, promotion in enumerate(map(partial(promote, promote_option=False),
                                      left_types,
                                      right_types)):
        if promotion == object_:
            raise TypeError(
                'Schemata of joining columns do not match,'
                ' no promotion found for %s=%s and %s=%s' % (
                    on_left[n], left_types[n], on_right[n], right_types[n],
                ),
            )
    _on_left = tuple(on_left) if isinstance(on_left, list) else on_left
    _on_right = (tuple(on_right) if isinstance(on_right, list)
                 else on_right)

    how = how.lower()
    if how not in ('inner', 'outer', 'left', 'right'):
        raise ValueError("How parameter should be one of "
                         "\n\tinner, outer, left, right."
                         "\nGot: %s" % how)

    return Join(lhs, rhs, _on_left, _on_right, how, suffixes)


class Concat(Expr):

    """ Stack tables on common columns

    Parameters
    ----------
    lhs, rhs : Expr
        Collections to concatenate
    axis : int, optional
        The axis to concatenate on.

    Examples
    --------
    >>> from blaze import symbol

    Vertically stack tables:

    >>> names = symbol('names', '5 * {name: string, id: int32}')
    >>> more_names = symbol('more_names', '7 * {name: string, id: int32}')
    >>> stacked = concat(names, more_names)
    >>> stacked.dshape
    dshape("12 * {name: string, id: int32}")

    Vertically stack matrices:

    >>> mat_a = symbol('a', '3 * 5 * int32')
    >>> mat_b = symbol('b', '3 * 5 * int32')
    >>> vstacked = concat(mat_a, mat_b, axis=0)
    >>> vstacked.dshape
    dshape("6 * 5 * int32")

    Horizontally stack matrices:

    >>> hstacked = concat(mat_a, mat_b, axis=1)
    >>> hstacked.dshape
    dshape("3 * 10 * int32")

    See Also
    --------

    blaze.expr.collections.Merge
    """
    _arguments = 'lhs', 'rhs', 'axis'
    __inputs__ = 'lhs', 'rhs'

    def _dshape(self):
        axis = self.axis
        ldshape = self.lhs.dshape
        lshape = ldshape.shape
        return DataShape(
            *(lshape[:axis] + (
                _shape_add(lshape[axis], self.rhs.dshape.shape[axis]),
            ) + lshape[axis + 1:] + (ldshape.measure,))
        )


def _shape_add(a, b):
    if isinstance(a, Var) or isinstance(b, Var):
        return var
    return Fixed(a.val + b.val)


@copydoc(Concat)
def concat(lhs, rhs, axis=0):
    ldshape = lhs.dshape
    rdshape = rhs.dshape
    if ldshape.measure != rdshape.measure:
        raise TypeError(
            'Mismatched measures: {l} != {r}'.format(
                l=ldshape.measure, r=rdshape.measure
            ),
        )

    lshape = ldshape.shape
    rshape = rdshape.shape
    for n, (a, b) in enumerate(zip_longest(lshape, rshape, fillvalue=None)):
        if n != axis and a != b:
            raise TypeError(
                'Shapes are not equal along axis {n}: {a} != {b}'.format(
                    n=n, a=a, b=b,
                ),
            )
    if axis < 0 or 0 < len(lshape) <= axis:
        raise ValueError(
            "Invalid axis '{a}', must be in range: [0, {n})".format(
                a=axis, n=len(lshape)
            ),
        )

    return Concat(lhs, rhs, axis)


class IsIn(ElemWise):
    """Check if an expression contains values from a set.

    Return a boolean expression indicating whether another expression
    contains values that are members of a collection.

    Parameters
    ----------
    expr : Expr
        Expression whose elements to check for membership in `keys`
    keys : Sequence
        Elements to test against. Blaze stores this as a ``frozenset``.

    Examples
    --------

    Check if a vector contains any of 1, 2 or 3:

    >>> from blaze import symbol
    >>> t = symbol('t', '10 * int64')
    >>> expr = t.isin([1, 2, 3])
    >>> expr.dshape
    dshape("10 * bool")
    """
    _arguments = '_child', '_keys'

    def _schema(self):
        return datashape.bool_

    def __str__(self):
        return '%s.%s(%s)' % (self._child, type(self).__name__.lower(),
                              self._keys)


@copydoc(IsIn)
def isin(expr, keys):
    if isinstance(keys, Expr):
        raise TypeError('keys argument cannot be an expression, '
                        'it must be an iterable object such as a list, '
                        'tuple or set')
    return IsIn(expr, frozenset(keys))


class Shift(Expr):
    """ Shift a column backward or forward by N elements

    Parameters
    ----------
    expr : Expr
        The expression to shift. This expression's dshape should be columnar
    n : int
        The number of elements to shift by. If n < 0 then shift backward,
        if n == 0 do nothing, else shift forward.
    """
    _arguments = '_child', 'n'

    def _schema(self):
        measure = self._child.schema.measure

        # if we are not shifting or we are already an Option type then return
        # the child's schema
        if not self.n or isinstance(measure, Option):
            return measure
        else:
            return Option(measure)

    def _dshape(self):
        return DataShape(*(self._child.dshape.shape + tuple(self.schema)))

    def __str__(self):
        return '%s(%s, n=%d)' % (
            type(self).__name__.lower(), self._child, self.n
        )


@copydoc(Shift)
def shift(expr, n):
    if not isinstance(n, (numbers.Integral, np.integer)):
        raise TypeError('n must be an integer')
    return Shift(expr, n)


dshape_method_list.extend([(iscollection, set([sort, head, tail, sample])),
                           (lambda ds: len(ds.shape) == 1, set([distinct, shift])),
                           (lambda ds: (len(ds.shape) == 1 and
                                        isscalar(getattr(ds.measure, 'key', ds.measure))), set([isin])),
                           (lambda ds: len(ds.shape) == 1 and isscalar(ds.measure), set([isin]))])



import pytest

import numpy as np
import pandas as pd
import pandas.util.testing as tm

dd = pytest.importorskip('dask.dataframe')

from blaze import symbol, discover, compute, broadcast_collect
from blaze.expr import mean, count, sum, min, max, var, std, summary


def eq(a, b):
    if isinstance(a, dd.DataFrame):
        tm.assert_frame_equal(a.compute(), b)
    elif isinstance(b, dd.Series):
        tm.assert_series_equal(a.compute(), b)


t = symbol('t', 'var * {name: string, amount: int, id: int}')
nt = symbol('t', 'var * {name: ?string, amount: float64, id: int}')

df = pd.DataFrame([['Alice', 100, 1],
                   ['Bob', 200, 2],
                   ['Alice', 50, 3]], columns=['name', 'amount', 'id'])

ndf = pd.DataFrame([['Alice', 100.0, 1],
                    ['Bob', np.nan, 2],
                    [np.nan, 50.0, 3]], columns=['name', 'amount', 'id'])

tbig = symbol('tbig',
              'var * {name: string, sex: string[1], amount: int, id: int}')

dfbig = pd.DataFrame([['Alice', 'F', 100, 1],
                      ['Alice', 'F', 100, 3],
                      ['Drew', 'F', 100, 4],
                      ['Drew', 'M', 100, 5],
                      ['Drew', 'M', 200, 5]],
                     columns=['name', 'sex', 'amount', 'id'])

ddf = dd.from_pandas(df, npartitions=2)
ddfbig = dd.from_pandas(dfbig, npartitions=2)


def test_broadcast():
    bcast = broadcast_collect(expr=t.amount * t.id)
    result = compute(bcast, ddf)
    eq(result, df.amount * df.id)


def test_symbol():
    eq(compute(t, ddf), df)


def test_sample():
    assert len(compute(t.sample(n=2), ddf)) == len(df.sample(n=2))
    assert len(compute(t.sample(frac=0.4), ddf)) == len(df.sample(frac=0.4))


def test_head():
    eq(compute(t.head(1), ddf), df.head(1))


def test_tail():
    eq(compute(t.tail(1), ddf), df.tail(1))


def test_relabel():
    result = compute(t.relabel({'name': 'NAME', 'id': 'ID'}), ddf)
    expected = df.rename(columns={'name': 'NAME', 'id': 'ID'})
    eq(result, expected)


def test_relabel_series():
    result = compute(t.relabel({'name': 'NAME'}), ddf.name)
    assert result.name == 'NAME'


def test_series_columnwise():
    data = pd.Series([1, 2, 3, 4], name='a')
    s = dd.from_pandas(data, npartitions=2)
    t = symbol('t', 'var * {a: int64}')
    result = compute(t.a + 1, s)
    eq(result, data)


def test_projection():
    eq(compute(t[['name', 'id']], ddf), df[['name', 'id']])


def test_field_on_series():
    expr = symbol('s', 'var * int')
    data = pd.Series([1, 2, 3, 4], name='s')
    s = dd.from_pandas(data, npartitions=2)
    eq(compute(expr.s, s), data)


def test_selection():
    eq(compute(t[t['amount'] == 0], ddf), df[df['amount'] == 0])
    eq(compute(t[t['amount'] > 150], ddf), df[df['amount'] > 150])


def test_selection_out_of_order():
    expr = t['name'][t['amount'] < 100]
    expected = df.loc[df.amount < 100, 'name']
    result = compute(expr, ddf)
    eq(result, expected)


def test_selection_inner_inputs():
    s_data = pd.DataFrame({'a': np.arange(5)})
    t_data = pd.DataFrame({'a': np.arange(5)})

    s_dd = dd.from_pandas(s_data, npartitions=2)
    t_dd = dd.from_pandas(t_data, npartitions=2)

    s = symbol('s', 'var * {a: int64}')
    t = symbol('t', 'var * {a: int64}')

    eq(compute(s[s.a == t.a], {s: s_dd, t: t_dd}), s_data)


def test_reductions():
    assert compute(mean(t['amount']), ddf) == 350 / 3
    assert compute(count(t['amount']), ddf) == 3
    assert compute(sum(t['amount']), ddf) == 100 + 200 + 50
    assert compute(min(t['amount']), ddf) == 50
    assert compute(max(t['amount']), ddf) == 200
    assert compute(var(t['amount']), ddf) == df.amount.var(ddof=0)
    assert compute(var(t['amount'], unbiased=True), ddf) == df.amount.var()
    assert compute(std(t['amount']), ddf) == df.amount.std(ddof=0)
    assert compute(std(t['amount'], unbiased=True), ddf) == df.amount.std()


def test_summary():
    expr = summary(count=t.id.count(), sum=t.amount.sum())
    eq(compute(expr, ddf), pd.Series({'count': 3, 'sum': 350}))


def test_summary_on_series():
    ser = dd.from_pandas(pd.Series([1, 2, 3]), npartitions=2)
    s = symbol('s', '3 * int')
    expr = summary(max=s.max(), min=s.min())
    assert compute(expr, ser) == (3, 1)

    expr = summary(max=s.max(), min=s.min(), keepdims=True)
    assert compute(expr, ser) == [(3, 1)]


@pytest.mark.parametrize('keys', [[1], [2, 3]])
def test_isin(keys):
    expr = t[t.id.isin(keys)]
    result = compute(expr, ddf)
    expected = df.loc[df.id.isin(keys)]
    eq(result, expected)


def test_coerce_series():
    s = pd.Series(list('1234'), name='a')
    dds = dd.from_pandas(s, npartitions=2)
    t = symbol('t', discover(s))
    result = compute(t.coerce(to='int64'), dds)
    expected = pd.Series([1, 2, 3, 4], name=s.name)
    eq(result, expected)


def test_nelements_count():
    assert compute(t.nelements(), ddf) == len(df)
    assert compute(t.nrows, ddf) == len(df)
    assert compute(count(t), ddf) == len(df)



from __future__ import absolute_import, division, print_function

default_map = map


def set_default_pmap(func):
    global default_map
    default_map = func


def get_default_pmap():
    return default_map



from __future__ import absolute_import, division, print_function
import threading

import numpy as np
from toolz import memoize
import datashape
import numba

from datashape import isdatelike, TimeDelta

from .core import optimize
from ..expr.strings import isstring
from ..expr import Expr, Map, BinOp, UnaryOp
from ..expr.broadcast import broadcast_collect, Broadcast
from .pyfunc import funcstr


Broadcastable = BinOp, Map, UnaryOp
lock = threading.Lock()


def optimize_ndarray(expr, *data, **kwargs):
    for leaf in expr._leaves():
        leaf_measure = leaf.dshape.measure

        # TODO: remove datelike skipping when numba/numba#1202 is fixed
        if (isstring(leaf_measure) or
            isdatelike(leaf_measure) or
            isinstance(leaf_measure, TimeDelta) or
            isinstance(leaf.dshape.measure, datashape.Record) and
            any(isstring(dt) or isdatelike(dt) or isinstance(dt, TimeDelta)
                for dt in leaf.dshape.measure.types)):
            return expr
    else:
        return broadcast_collect(
            expr,
            broadcastable=Broadcastable,
            want_to_broadcast=Broadcastable,
        )


for i in range(1, 11):
    optimize.register(Expr, *([np.ndarray] * i))(optimize_ndarray)


def get_numba_type(dshape):
    """Get the numba type corresponding to the ``datashape.Mono`` instance.
    `dshape`

    Parameters
    ----------
    dshape : datashape.Mono

    Returns
    -------
    restype : numba.types.Type
        Numba type corresponding to `dshape`.

    Examples
    --------
    >>> get_numba_type(datashape.bool_) == numba.bool_
    True

    See Also
    --------
    compute_signature
    """
    measure = dshape.measure
    if measure == datashape.bool_:
        restype = numba.bool_  # str(bool_) == 'bool' so we can't use getattr
    elif measure == datashape.date_:
        restype = numba.types.NPDatetime('D')
    elif measure == datashape.datetime_:
        restype = numba.types.NPDatetime('us')
    elif isinstance(measure, datashape.TimeDelta):  # isinstance for diff freqs
        restype = numba.types.NPTimedelta(measure.unit)
    elif isinstance(measure, datashape.String):
        encoding = measure.encoding
        fixlen = measure.fixlen
        if fixlen is None:
            if encoding == 'A':
                return numba.types.string
            raise TypeError("Numba cannot handle variable length strings")
        typ = (numba.types.CharSeq
               if encoding == 'A' else numba.types.UnicodeCharSeq)
        return typ(fixlen or 0)
    elif measure == datashape.object_:
        raise TypeError("Numba cannot handle object datashape")
    else:
        try:
            restype = getattr(numba, str(measure))
        except AttributeError:
            raise TypeError('Invalid datashape to numba type: %r' % measure)
    return restype


def compute_signature(expr):
    """Get the ``numba`` *function signature* corresponding to ``DataShape``

    Examples
    --------
    >>> from blaze import symbol
    >>> s = symbol('s', 'int64')
    >>> t = symbol('t', 'float32')
    >>> from numba import float64, int64, float32
    >>> expr = s + t
    >>> compute_signature(expr) == float64(int64, float32)
    True

    Notes
    -----
    * This could potentially be adapted/refactored to deal with
      ``datashape.Function`` types.
    * Cannot handle ``datashape.Record`` types.
    """
    assert datashape.isscalar(expr.schema)
    restype = get_numba_type(expr.schema)
    argtypes = [get_numba_type(e.schema) for e in expr._leaves()]
    return restype(*argtypes)


def _get_numba_ufunc(expr):
    """Construct a numba ufunc from a blaze expression

    Parameters
    ----------
    expr : blaze.expr.Expr

    Returns
    -------
    f : function
        A numba vectorized function

    Examples
    --------
    >>> from blaze import symbol
    >>> import numpy as np

    >>> s = symbol('s', 'float64')
    >>> t = symbol('t', 'float64')

    >>> x = np.array([1.0, 2.0, 3.0])
    >>> y = np.array([2.0, 3.0, 4.0])

    >>> f = get_numba_ufunc(s + t)

    >>> f(x, y)
    array([ 3.,  5.,  7.])

    See Also
    --------
    get_numba_type
    compute_signature
    """
    if isinstance(expr, Broadcast):
        leaves = expr._scalars
        expr = expr._scalar_expr
    else:
        leaves = expr._leaves()

    s, scope = funcstr(leaves, expr)

    scope = dict((k, numba.jit(nopython=True)(v) if callable(v) else v)
                 for k, v in scope.items())
    # get the func
    func = eval(s, scope)
    # get the signature
    sig = compute_signature(expr)
    # vectorize is currently not thread safe. So lock the thread.
    # TODO FIXME remove this when numba has made vectorize thread safe.
    with lock:
        ufunc = numba.vectorize([sig], nopython=True)(func)
    return ufunc


# do this here so we can run our doctest
get_numba_ufunc = memoize(_get_numba_ufunc)


def broadcast_numba(t, *data, **kwargs):
    return get_numba_ufunc(t)(*data)



import blaze

def test_basic_interface():
    blaze.compute
    blaze.into
    blaze.discover

def test_no_leeking():
    # unique is a commonly used toolz function and a good example of something
    # that shouldn't leak out to the main namespace
    assert not hasattr(blaze, 'unique')



import pytest

pytest.importorskip('sqlalchemy')

from functools import partial

from toolz import first
from sqlalchemy.exc import OperationalError
from odo import into, drop

from blaze import data as bz_data
from blaze import create_index


@pytest.fixture
def sql():
    data = [(1, 2), (10, 20), (100, 200)]
    sql = bz_data(
        'sqlite:///:memory:::foo',
        dshape='var * {x: int, y: int}',
    )
    into(sql, data)
    return sql


def test_column(sql):
    t = bz_data(sql)

    r = list(t['x'])
    assert r == [1, 10, 100]
    assert list(t[['x']]) == [(1,), (10,), (100,)]

    assert int(t.count()) == 3


def test_drop(sql):
    sql = sql.data
    assert sql.exists(sql.bind)
    drop(sql)
    assert not sql.exists(sql.bind)


@pytest.mark.parametrize('cols', (
    'x', ['x'], ['y'], ['x', 'y'], ('x',), ('y',), ('x', 'y'),
))
def test_create_index(sql, cols):
    create_index(sql, cols, name='idx')
    with pytest.raises(OperationalError):
        create_index(sql, cols, name='idx')


def test_create_index_fails(sql):
    with pytest.raises(KeyError):
        create_index(sql, 'z', name='zidx')
    with pytest.raises(ValueError):
        create_index(sql, 'x')
    with pytest.raises(ValueError):
        create_index(sql, 'z')


def test_create_index_unique(sql):
    create_index(sql, 'y', name='y_idx', unique=True)
    assert len(sql.data.indexes) == 1
    idx = first(sql.data.indexes)
    assert idx.unique
    assert idx.columns.y == sql.data.c.y


def test_composite_index(sql):
    create_index(sql, ['x', 'y'], name='idx_xy')
    with pytest.raises(OperationalError):
        create_index(sql, ['x', 'y'], name='idx_xy')


def test_composite_index_fails(sql):
    with pytest.raises(KeyError):
        create_index(sql, ['z', 'bizz'], name='idx_name')


def test_composite_index_fails_with_existing_columns(sql):
    with pytest.raises(KeyError):
        create_index(sql, ['x', 'z', 'bizz'], name='idx_name')


@pytest.mark.parametrize('cols', ('x', ['x', 'y']))
def test_ignore_existing(sql, cols):
    create_call = partial(
        create_index,
        sql,
        cols,
        name='idx_name',
    )

    create_call()
    with pytest.raises(OperationalError):
        create_call(ignore_existing=False)

    # Shouldn't error
    create_call(ignore_existing=True)



from blaze.expr.optimize import lean_projection, _lean
from blaze.expr import symbol, summary, by, By, Field, Projection, merge, sum

t = symbol('t', 'var * {x: int, y: int, z: int, w: int}')


def test_lean_on_Symbol():
    assert _lean(t, fields=['x'])[0] == t[['x']]
    assert _lean(t, fields=['x', 'y', 'z', 'w'])[0] == t


def test_lean_projection():
    assert lean_projection(t[t.x > 0].y)._child._child.isidentical(
        t[['x', 'y']],
    )


def test_lean_projection_by():
    assert lean_projection(by(t.x, total=t.y.sum()))._child.isidentical(
        t[['x', 'y']],
    )


def test_lean_by_with_summary():
    assert lean_projection(by(t.x, total=t.y.sum()))._child.isidentical(
        t[['x', 'y']],
    )

    tt = t[['x', 'y']]
    result = lean_projection(by(t.x, a=t.y.sum(), b=t.z.sum())[['x', 'a']])
    expected = Projection(
        By(Field(tt, 'x'), summary(a=sum(Field(tt, 'y')))),
        ('x', 'a'),
    )
    assert result.isidentical(expected)


def test_summary():
    expr, fields = _lean(summary(a=t.x.sum(), b=t.y.sum()), fields=['a'])
    assert expr.isidentical(summary(a=t.x.sum()))
    assert fields == {'x'}


def test_sort():
    assert lean_projection(t.sort('x').y).isidentical(t[['x','y']].sort('x').y)


def test_merge():
    expr = lean_projection(merge(a=t.x+1, y=t.y))
    assert expr._child.isidentical(t[['x', 'y']])


def test_add():
    expr = t.x + 1
    expr2, fields = _lean(expr, fields=set(['x']))
    assert expr2.isidentical(expr)
    assert fields == {'x'}

    expr = (t.x + t.y).label('a')
    expr2, fields = _lean(expr, fields=set(['a']))
    assert expr2.isidentical(expr)
    assert fields == {'x', 'y'}


def test_label():
    expr = t.x.label('foo')
    expr2, fields =  _lean(expr, fields=set(['foo']))
    assert expr2.isidentical(expr)
    assert fields == {'x'}


def test_relabel():
    expr = t.relabel(x='X', y='Y')
    expr2, fields = _lean(expr, fields=set(['X', 'z']))
    assert expr2.isidentical(t[['x', 'z']].relabel(x='X'))
    assert fields == {'x', 'z'}


def test_merge_with_table():
    expr = lean_projection(merge(t, a=t.x+1))
    assert expr.isidentical(expr)  # wut?


def test_head():
    assert lean_projection(t.sort('x').y.head(5)).isidentical(
        t[['x', 'y']].sort('x').y.head(5),
    )


def test_elemwise_thats_also_a_column():
    t = symbol('t', 'var * {x: int, time: datetime, y: int}')
    expr = t[t.x > 0].time.truncate(months=1)
    expected = t[['time', 'x']]
    result = lean_projection(expr)
    assert result._child._child._child.isidentical(expected)


def test_distinct():
    expr = t.distinct()[['x', 'y']]
    assert lean_projection(expr).isidentical(expr)


def test_like():
    t = symbol('t', 'var * {name: string, x: int, y: int}')
    expr = t[t.name.like('Alice')].y
    result = lean_projection(expr)
    assert result._child._child.isidentical(t[['name', 'y']])



from __future__ import absolute_import, division, print_function

from .server import Server, to_tree, from_tree, api, expr_md5
from .spider import data_spider, from_yaml
from .client import Client
from .serialization import (
    SerializationFormat,
    all_formats,
    fastmsgpack as fastmsgpack_format,
    json as json_format,
    msgpack as msgpack_format,
    pickle as pickle_format,
)


__all__ = [
    'Client',
    'SerializationFormat',
    'Server',
    'all_formats',
    'api',
    'data_spider',
    'expr_md5',
    'fastmsgpack_format',
    'from_tree',
    'from_yaml',
    'json_format',
    'msgpack_format',
    'pickle_format',
    'to_tree',
]



from blaze.expr import *
from blaze.expr.broadcast import *
from blaze.expr.broadcast import leaves_of_type, broadcast_collect
from blaze.compatibility import builtins
from toolz import isdistinct

x = symbol('x', '5 * 3 * int32')
xx = symbol('xx', 'int32')

y = symbol('y', '5 * 3 * int32')
yy = symbol('yy', 'int32')

a = symbol('a', 'int32')


def test_broadcast_basic():
    b = Broadcast((x, y), (xx, yy), xx + yy)

    assert b.shape == x.shape
    assert b.schema == (xx + yy).dshape

    assert eval(str(b)).isidentical(b)


def test_scalar_symbols():
    exprs = [x, y]
    scalars = scalar_symbols(exprs)

    assert len(scalars) == len(exprs)
    assert isdistinct([s._name for s in scalars])
    assert builtins.all(s.dshape == e.schema for s, e in zip(scalars, exprs))


def test_broadcast_function():
    expr =  Pow(Add(x, Mult(2, y)), 2)  # (x + (2 * y)) ** 2
    b = broadcast(expr, [x, y])
    xx, yy = b._scalars
    assert b._scalar_expr.isidentical((xx + (2 * yy)) ** 2)

    # A different set of leaves
    b = broadcast(expr, [x, Mult(2, y)])
    xx, yy = b._scalars
    assert b._scalar_expr.isidentical((xx + yy) ** 2)


t = symbol('t', 'var * {x: int, y: int, z: int}')


def test_tabular_case():
    expr = t.x + t.y * 2

    b = broadcast(expr, [t])
    tt, = b._scalars

    assert b._scalar_expr.isidentical(tt.x + tt.y * 2)


def test_optimize_broadcast():
    expr = (t.distinct().x + 1).distinct()

    expected = broadcast(t.distinct().x + 1, [t.distinct()]).distinct()
    result = broadcast_collect(
        expr,
        broadcastable=(Field, Arithmetic),
        want_to_broadcast=(Field, Arithmetic),
    )

    assert result.isidentical(expected)


def test_leaves_of_type():
    expr = Distinct(Distinct(Distinct(t.x)))

    result = leaves_of_type((Distinct,), expr)
    assert len(result) == 1
    assert list(result)[0].isidentical(t.x)


def test_broadcast_collect_doesnt_collect_scalars():
    expr = xx + yy * a

    assert broadcast_collect(expr,
                             broadcastable=Arithmetic,
                             want_to_broadcast=Arithmetic).isidentical(expr)


def test_table_broadcast():
    t = symbol('t', 'var * {x: int, y: int, z: int}')

    expr = t.distinct()
    expr = (2 * expr.x + expr.y + 1).distinct()

    expected = t.distinct()
    expected = broadcast(2 * expected.x + expected.y + 1, [expected]).distinct()
    assert broadcast_collect(expr).isidentical(expected)

    expr = (t.x + t.y).sum()
    result = broadcast_collect(expr)
    expected = broadcast(t.x + t.y, [t]).sum()
    assert result.isidentical(expected)


def test_broadcast_doesnt_affect_scalars():
    t = symbol('t', '{x: int, y: int, z: int}')
    expr = (2 * t.x + t.y + 1)

    assert broadcast_collect(expr).isidentical(expr)


def test_full_expr():
    b = Broadcast((x, y), (xx, yy), xx + yy)
    assert b._full_expr.isidentical(x + y)


def test_broadcast_naming():
    t = symbol('t', 'var * {x: int, y: int, z: int}')

    for expr in [t.x, t.x + 1]:
        assert broadcast(expr, [t])._name == 'x'



from __future__ import absolute_import, division, print_function

import pytest

from datetime import datetime, timedelta, date, time

import numpy as np

import pandas as pd
import pandas.util.testing as tm
from pandas import DataFrame, Series

from string import ascii_lowercase

from blaze.compute.core import compute
from blaze.compute.pandas import pdsort
from blaze import dshape, discover, transform, broadcast_collect
import datashape
from blaze.expr import symbol, join, by, summary, distinct, shape
from blaze.expr import (merge, exp, mean, count, nunique, sum, min, max, any,
                        var, std, concat, coalesce)
from blaze.compatibility import builtins, xfail, assert_series_equal
import sys

t = symbol('t', 'var * {name: string, amount: int, id: int}')
nt = symbol('t', 'var * {name: ?string, amount: float64, id: int}')


df = DataFrame([['Alice', 100, 1],
                ['Bob', 200, 2],
                ['Alice', 50, 3]], columns=['name', 'amount', 'id'])

ndf = DataFrame([['Alice', 100.0, 1],
                 ['Bob', np.nan, 2],
                 [np.nan, 50.0, 3]], columns=['name', 'amount', 'id'])


tbig = symbol('tbig',
              'var * {name: string, sex: string[1], amount: int, id: int}')

dfbig = DataFrame([['Alice', 'F', 100, 1],
                   ['Alice', 'F', 100, 3],
                   ['Drew', 'F', 100, 4],
                   ['Drew', 'M', 100, 5],
                   ['Drew', 'M', 200, 5]],
                  columns=['name', 'sex', 'amount', 'id'])


# for now jsut copy this, but will open a PR to see if we can remove some of
# the repetitive copying
tbgr = symbol('tbgr',
              """ var * {name: string,
                         sex: string[1],
                         amount: int,
                         id: int,
                         comment: ?string}
              """)

dfbgr = DataFrame([['Alice', 'F', 100, 1, 'Alice comment'],
                   ['Alice', 'F', 100, 3, None],
                   ['Drew', 'F', 100, 4, 'Drew comment'],
                   ['Drew', 'M', 100, 5, 'Drew comment 2'],
                   ['Drew', 'M', 200, 5, None]],
                  columns=['name', 'sex', 'amount', 'id', 'comment'])


@pytest.fixture(scope='module')
def df_add_null():
    rows = [(None, 'M', 300, 6),
            ('first', None, 300, 6),
            (None, None, 300, 6)]
    df_add_null = dfbig.append(DataFrame(rows,
                                         columns=dfbig.columns
                                         ), ignore_index=True)

    return df_add_null


def test_series_broadcast():
    s = Series([1, 2, 3], name='a')
    t = symbol('t', 'var * {a: int64}')
    bcast = broadcast_collect(expr=(2 * t.a - t.a**2))
    result = compute(bcast, s)
    assert_series_equal(result, 2 * s - s**2)


def test_frame_broadcast():
    bcast = broadcast_collect(expr=t.amount * t.id)
    result = compute(bcast, df)
    assert_series_equal(result, df.amount * df.id)


def test_series_columnwise():
    s = Series([1, 2, 3], name='a')
    t = symbol('t', 'var * {a: int64}')
    result = compute(t.a + 1, s)
    assert_series_equal(s + 1, result)


def test_symbol():
    tm.assert_frame_equal(compute(t, df), df)


def test_projection():
    tm.assert_frame_equal(compute(t[['name', 'id']], df),
                          df[['name', 'id']])


def test_eq():
    assert_series_equal(compute(t['amount'] == 100, df),
                           df['amount'] == 100)


def test_selection():
    tm.assert_frame_equal(compute(t[t['amount'] == 0], df),
                          df[df['amount'] == 0])
    tm.assert_frame_equal(compute(t[t['amount'] > 150], df),
                          df[df['amount'] > 150])


def test_arithmetic():
    assert_series_equal(compute(t['amount'] + t['id'], df),
                           df.amount + df.id)
    assert_series_equal(compute(t['amount'] * t['id'], df),
                           df.amount * df.id)
    assert_series_equal(compute(t['amount'] % t['id'], df),
                           df.amount % df.id)


def test_join():
    left = DataFrame(
        [['Alice', 100], ['Bob', 200]], columns=['name', 'amount'])
    right = DataFrame([['Alice', 1], ['Bob', 2]], columns=['name', 'id'])

    lsym = symbol('L', 'var * {name: string, amount: int}')
    rsym = symbol('R', 'var * {name: string, id: int}')
    joined = join(lsym, rsym, 'name')

    assert (dshape(joined.schema) ==
            dshape('{name: string, amount: int, id: int}'))

    result = compute(joined, {lsym: left, rsym: right})

    expected = DataFrame([['Alice', 100, 1], ['Bob', 200, 2]],
                         columns=['name', 'amount', 'id'])

    tm.assert_frame_equal(result, expected)
    assert list(result.columns) == list(joined.fields)


def test_multi_column_join():
    left = [(1, 2, 3),
            (2, 3, 4),
            (1, 3, 5)]
    left = DataFrame(left, columns=['x', 'y', 'z'])
    right = [(1, 2, 30),
             (1, 3, 50),
             (1, 3, 150)]
    right = DataFrame(right, columns=['x', 'y', 'w'])

    lsym = symbol('lsym', 'var * {x: int, y: int, z: int}')
    rsym = symbol('rsym', 'var * {x: int, y: int, w: int}')

    j = join(lsym, rsym, ['x', 'y'])

    expected = [(1, 2, 3, 30),
                (1, 3, 5, 50),
                (1, 3, 5, 150)]
    expected = DataFrame(expected, columns=['x', 'y', 'z', 'w'])

    result = compute(j, {lsym: left, rsym: right})

    print(result)

    tm.assert_frame_equal(result, expected)
    assert list(result.columns) == list(j.fields)


def test_unary_op():
    assert (compute(exp(t['amount']), df) == np.exp(df['amount'])).all()


def test_abs():
    assert (compute(abs(t['amount']), df) == abs(df['amount'])).all()


def test_neg():
    assert_series_equal(compute(-t['amount'], df),
                           -df['amount'])


@xfail(reason='Projection does not support arithmetic')
def test_neg_projection():
    assert_series_equal(compute(-t[['amount', 'id']], df),
                           -df[['amount', 'id']])


def test_columns_series():
    assert isinstance(compute(t['amount'], df), Series)
    assert isinstance(compute(t['amount'] > 150, df), Series)


def test_reductions():
    assert compute(mean(t['amount']), df) == 350 / 3
    assert compute(count(t['amount']), df) == 3
    assert compute(sum(t['amount']), df) == 100 + 200 + 50
    assert compute(min(t['amount']), df) == 50
    assert compute(max(t['amount']), df) == 200
    assert compute(nunique(t['amount']), df) == 3
    assert compute(nunique(t['name']), df) == 2
    assert compute(any(t['amount'] > 150), df) is True
    assert compute(any(t['amount'] > 250), df) is False
    assert compute(var(t['amount']), df) == df.amount.var(ddof=0)
    assert compute(var(t['amount'], unbiased=True), df) == df.amount.var()
    assert compute(std(t['amount']), df) == df.amount.std(ddof=0)
    assert compute(std(t['amount'], unbiased=True), df) == df.amount.std()
    assert compute(t.amount[0], df) == df.amount.iloc[0]
    assert compute(t.amount[-1], df) == df.amount.iloc[-1]


def test_reductions_on_dataframes():
    assert compute(count(t), df) == 3
    assert shape(compute(count(t, keepdims=True), df)) == (1,)


def test_1d_reductions_keepdims():
    series = df['amount']
    for r in [sum, min, max, nunique, count, std, var]:
        result = compute(r(t.amount, keepdims=True), {t.amount: series})
        assert type(result) == type(series)


def test_distinct():
    dftoobig = DataFrame([['Alice', 'F', 100, 1],
                          ['Alice', 'F', 100, 1],
                          ['Alice', 'F', 100, 3],
                          ['Drew', 'F', 100, 4],
                          ['Drew', 'M', 100, 5],
                          ['Drew', 'F', 100, 4],
                          ['Drew', 'M', 100, 5],
                          ['Drew', 'M', 200, 5],
                          ['Drew', 'M', 200, 5]],
                         columns=['name', 'sex', 'amount', 'id'])
    d_t = distinct(tbig)
    d_df = compute(d_t, dftoobig)
    tm.assert_frame_equal(d_df, dfbig)
    # Test idempotence
    tm.assert_frame_equal(compute(d_t, d_df), d_df)


def test_distinct_on():
    cols = ['name', 'sex', 'amount', 'id']
    df = DataFrame([['Alice', 'F', 100, 1],
                    ['Alice', 'F', 100, 3],
                    ['Drew', 'F', 100, 4],
                    ['Drew', 'M', 100, 5],
                    ['Drew', 'F', 100, 4],
                    ['Drew', 'M', 100, 5],
                    ['Drew', 'M', 200, 5]],
                   columns=cols)
    s = symbol('s', discover(df))
    computed = compute(s.distinct('sex'), df)
    tm.assert_frame_equal(
        computed,
        pd.DataFrame([['Alice', 'F', 100, 1],
                      ['Drew', 'M', 100, 5]],
                     columns=cols),
    )


def test_by_one():
    result = compute(by(t['name'], total=t['amount'].sum()), df)
    expected = df.groupby('name')['amount'].sum().reset_index()
    expected.columns = ['name', 'total']
    tm.assert_frame_equal(result, expected)


def test_by_two():
    result = compute(by(tbig[['name', 'sex']],
                        total=sum(tbig['amount'])), dfbig)

    expected = DataFrame([['Alice', 'F', 200],
                          ['Drew', 'F', 100],
                          ['Drew', 'M', 300]],
                         columns=['name', 'sex', 'total'])

    tm.assert_frame_equal(result, expected)


def test_by_three():

    expr = by(tbig[['name', 'sex']],
              total=(tbig['id'] + tbig['amount']).sum())

    result = compute(expr, dfbig)

    expected = DataFrame([['Alice', 'F', 204],
                          ['Drew', 'F', 104],
                          ['Drew', 'M', 310]], columns=['name', 'sex', 'total'])
    expected.columns = expr.fields
    tm.assert_frame_equal(result, expected)


def test_by_four():
    t = tbig[['sex', 'amount']]
    expr = by(t['sex'], max=t['amount'].max())
    result = compute(expr, dfbig)

    expected = DataFrame([['F', 100],
                          ['M', 200]], columns=['sex', 'max'])

    tm.assert_frame_equal(result, expected)


def test_join_by_arcs():
    df_idx = DataFrame([['A', 1],
                        ['B', 2],
                        ['C', 3]],
                       columns=['name', 'node_id'])

    df_arc = DataFrame([[1, 3],
                        [2, 3],
                        [3, 1]],
                       columns=['node_out', 'node_id'])

    t_idx = symbol('t_idx', 'var * {name: string, node_id: int32}')

    t_arc = symbol('t_arc', 'var * {node_out: int32, node_id: int32}')

    joined = join(t_arc, t_idx, "node_id")

    want = by(joined['name'], count=joined['node_id'].count())

    result = compute(want, {t_arc: df_arc, t_idx: df_idx})

    result_pandas = pd.merge(df_arc, df_idx, on='node_id')

    gb = result_pandas.groupby('name')
    expected = gb.node_id.count().reset_index().rename(columns={
                                                       'node_id': 'count'
                                                       })

    tm.assert_frame_equal(result, expected)
    assert list(result.columns) == ['name', 'count']


def test_join_suffixes():
    df = pd.DataFrame(
        list(dict((k, n) for k in ascii_lowercase[:5]) for n in range(5)),
    )
    a = symbol('a', discover(df))
    b = symbol('b', discover(df))

    suffixes = '_x', '_y'
    joined = join(a, b, 'a', suffixes=suffixes)

    expected = pd.merge(df, df, on='a', suffixes=suffixes)
    result = compute(joined, {a: df, b: df})
    tm.assert_frame_equal(result, expected)


def test_join_promotion():
    a_data = pd.DataFrame([[0.0, 1.5], [1.0, 2.5]], columns=list('ab'))
    b_data = pd.DataFrame([[0, 1], [1, 2]], columns=list('ac'))
    a = symbol('a', discover(a_data))
    b = symbol('b', discover(b_data))

    joined = join(a, b, 'a')
    assert joined.dshape == dshape('var * {a: float64, b: float64, c: int64}')

    expected = pd.merge(a_data, b_data, on='a')
    result = compute(joined, {a: a_data, b: b_data})
    tm.assert_frame_equal(result, expected)


def test_sort():
    tm.assert_frame_equal(compute(t.sort('amount'), df),
                          pdsort(df, 'amount'))

    tm.assert_frame_equal(compute(t.sort('amount', ascending=True), df),
                          pdsort(df, 'amount', ascending=True))

    tm.assert_frame_equal(compute(t.sort(['amount', 'id']), df),
                          pdsort(df, ['amount', 'id']))


def test_sort_on_series_no_warning(recwarn):
    expected = df.amount.order()

    recwarn.clear()

    assert_series_equal(compute(t['amount'].sort('amount'), df), expected)

    # raises as assertion error if no warning occurs, same thing for below
    with pytest.raises(AssertionError):
        assert recwarn.pop(FutureWarning)

    assert_series_equal(compute(t['amount'].sort(), df), expected)
    with pytest.raises(AssertionError):
        assert recwarn.pop(FutureWarning)


def test_field_on_series():
    expr = symbol('s', 'var * int')
    data = Series([1, 2, 3, 4], name='s')
    assert_series_equal(compute(expr.s, data), data)


def test_head():
    tm.assert_frame_equal(compute(t.head(1), df), df.head(1))


def test_tail():
    tm.assert_frame_equal(compute(t.tail(1), df), df.tail(1))


def test_sample():
    samp = compute(t.sample(n=2), df)
    assert len(samp) == 2


def test_sample_frac_rounding_edge_case():
    samp_big = compute(tbig.sample(frac=0.1), dfbig)
    assert len(samp_big) == int(np.ceil(len(dfbig) * 0.1))


def test_sample_clip():
    samp_series = compute(t.name.sample(n=2*len(df)), df)
    samp_df = compute(t.sample(n=2*len(df)), df)
    assert len(samp_series) == len(samp_df) == len(df)


def test_label():
    expected = df['amount'] * 10
    expected.name = 'foo'
    assert_series_equal(compute((t['amount'] * 10).label('foo'), df),
                           expected)


def test_relabel():
    result = compute(t.relabel({'name': 'NAME', 'id': 'ID'}), df)
    expected = df.rename(columns={'name': 'NAME', 'id': 'ID'})
    tm.assert_frame_equal(result, expected)


def test_relabel_series():
    result = compute(t.relabel({'name': 'NAME'}), df.name)
    assert result.name == 'NAME'


ts = pd.date_range('now', periods=10).to_series().reset_index(drop=True)
tframe = DataFrame({'timestamp': ts})


def test_map_column():
    inc = lambda x: x + 1
    result = compute(t['amount'].map(inc, 'int'), df)
    expected = df['amount'] + 1
    assert_series_equal(result, expected)


def test_map():
    f = lambda _, amt, id: amt + id
    result = compute(t.map(f, 'real'), df)
    expected = df['amount'] + df['id']
    assert_series_equal(result, expected)


def test_apply_column():
    result = compute(t.amount.apply(np.sum, 'real'), df)
    expected = np.sum(df['amount'])
    assert result == expected

    result = compute(t.amount.apply(builtins.sum, 'real'), df)
    expected = builtins.sum(df['amount'])
    assert result == expected


def test_apply():
    result = compute(t.apply(str, 'string'), df)
    expected = str(df)
    assert result == expected


def test_merge():
    col = (t['amount'] * 2).label('new')
    expr = merge(t['name'], col)

    expected = DataFrame([['Alice', 200],
                          ['Bob', 400],
                          ['Alice', 100]],
                         columns=['name', 'new'])

    result = compute(expr, df)
    tm.assert_frame_equal(result, expected)


def test_by_nunique():
    result = compute(by(t['name'], count=t['id'].nunique()), df)
    expected = DataFrame([['Alice', 2], ['Bob', 1]],
                         columns=['name', 'count'])

    tm.assert_frame_equal(result, expected)


def test_selection_out_of_order():
    expr = t['name'][t['amount'] < 100]
    expected = df.loc[df.amount < 100, 'name']
    result = compute(expr, df)
    assert_series_equal(result, expected)


def test_outer_join():
    left = [(1, 'Alice', 100),
            (2, 'Bob', 200),
            (4, 'Dennis', 400)]
    left = DataFrame(left, columns=['id', 'name', 'amount'])

    right = [('NYC', 1),
             ('Boston', 1),
             ('LA', 3),
             ('Moscow', 4)]
    right = DataFrame(right, columns=['city', 'id'])

    lsym = symbol('lsym', 'var * {id: int, name: string, amount: real}')
    rsym = symbol('rsym', 'var * {city: string, id: int}')

    convert = lambda df: set(df.to_records(index=False).tolist())

    assert (convert(compute(join(lsym, rsym), {lsym: left, rsym: right})) ==
            set([(1, 'Alice', 100, 'NYC'),
                 (1, 'Alice', 100, 'Boston'),
                 (4, 'Dennis', 400, 'Moscow')]))

    assert (convert(compute(join(lsym, rsym, how='left'),
                            {lsym: left, rsym: right})) ==
            set([(1, 'Alice', 100, 'NYC'),
                 (1, 'Alice', 100, 'Boston'),
                 (2, 'Bob', 200, np.nan),
                 (4, 'Dennis', 400, 'Moscow')]))

    df = compute(join(lsym, rsym, how='right'), {lsym: left, rsym: right})
    expected = DataFrame([(1., 'Alice', 100., 'NYC'),
                          (1., 'Alice', 100., 'Boston'),
                          (3., np.nan, np.nan, 'lsymA'),
                          (4., 'Dennis', 400., 'Moscow')],
                         columns=['id', 'name', 'amount', 'city'])

    result = pdsort(df, 'id').to_records(index=False)
    expected = pdsort(expected, 'id').to_records(index=False)
    np.array_equal(result, expected)

    df = compute(join(lsym, rsym, how='outer'), {lsym: left, rsym: right})
    expected = DataFrame([(1., 'Alice', 100., 'NYC'),
                          (1., 'Alice', 100., 'Boston'),
                          (2., 'Bob', 200., np.nan),
                          (3., np.nan, np.nan, 'LA'),
                          (4., 'Dennis', 400., 'Moscow')],
                         columns=['id', 'name', 'amount', 'city'])

    result = pdsort(df, 'id').to_records(index=False)
    expected = pdsort(expected, 'id').to_records(index=False)
    np.array_equal(result, expected)


def test_by_on_same_column():
    df = pd.DataFrame([[1, 2], [1, 4], [2, 9]], columns=['id', 'value'])
    t = symbol('data', 'var * {id: int, value: int}')

    gby = by(t['id'], count=t['id'].count())

    expected = DataFrame([[1, 2], [2, 1]], columns=['id', 'count'])
    result = compute(gby, {t: df})
    tm.assert_frame_equal(result, expected)


def test_summary_by():
    expr = by(t.name, summary(count=t.id.count(), sum=t.amount.sum()))
    result = compute(expr, df)
    expected = DataFrame([['Alice', 2, 150],
                          ['Bob', 1, 200]], columns=['name', 'count', 'sum'])

    expr = by(t.name, summary(count=t.id.count(), sum=(t.amount + 1).sum()))
    result = compute(expr, df)
    expected = DataFrame([['Alice', 2, 152],
                          ['Bob', 1, 201]], columns=['name', 'count', 'sum'])
    tm.assert_frame_equal(result, expected)

@pytest.mark.xfail(raises=TypeError,
                   reason=('pandas backend cannot support non Reduction '
                           'subclasses'))
def test_summary_by_first():
    expr = by(t.name, fst=t.amount[0])
    result = compute(expr, df)
    assert result == df.amount.iloc[0]


def test_summary_by_reduction_arithmetic():
    expr = by(t.name, summary(count=t.id.count(), sum=t.amount.sum() + 1))
    result = compute(expr, df)
    expected = DataFrame([['Alice', 2, 151],
                          ['Bob', 1, 201]], columns=['name', 'count', 'sum'])
    tm.assert_frame_equal(result, expected)


def test_summary():
    expr = summary(count=t.id.count(), sum=t.amount.sum())
    assert_series_equal(compute(expr, df), Series({'count': 3, 'sum': 350}))


def test_summary_on_series():
    ser = Series([1, 2, 3])
    s = symbol('s', '3 * int')
    expr = summary(max=s.max(), min=s.min())
    assert compute(expr, ser) == (3, 1)

    expr = summary(max=s.max(), min=s.min(), keepdims=True)
    assert compute(expr, ser) == [(3, 1)]


def test_summary_keepdims():
    expr = summary(count=t.id.count(), sum=t.amount.sum(), keepdims=True)
    expected = DataFrame([[3, 350]], columns=['count', 'sum'])
    tm.assert_frame_equal(compute(expr, df), expected)


def test_dplyr_transform():
    df = DataFrame({'timestamp': pd.date_range('now', periods=5)})
    t = symbol('t', discover(df))
    expr = transform(t, date=t.timestamp.map(lambda x: x.date(),
                                             schema='datetime'))
    lhs = compute(expr, df)
    rhs = pd.concat([df, Series(df.timestamp.map(lambda x: x.date()),
                                name='date').to_frame()], axis=1)
    tm.assert_frame_equal(lhs, rhs)


def test_nested_transform():
    d = {'timestamp': [1379613528, 1379620047], 'platform': ["Linux",
                                                             "Windows"]}
    df = DataFrame(d)
    t = symbol('t', discover(df))
    t = transform(t, timestamp=t.timestamp.map(datetime.fromtimestamp,
                                               schema='datetime'))
    expr = transform(t, date=t.timestamp.map(lambda x: x.date(),
                                             schema='datetime'))
    result = compute(expr, df)
    df['timestamp'] = df.timestamp.map(datetime.fromtimestamp)
    df['date'] = df.timestamp.map(lambda x: x.date())
    tm.assert_frame_equal(result, df)


def test_transform_with_common_subexpression():
    df = DataFrame(np.random.rand(5, 2), columns=list('ab'))
    t = symbol('t', discover(df))
    expr = transform(t, c=t.a - t.a % 3, d=t.a % 3)
    result = compute(expr, df)
    expected = pd.concat(
        [df[c] for c in df.columns] + [
            pd.Series(df.a - df.a % 3, name='c'),
            pd.Series(df.a % 3, name='d')
        ],
        axis=1
    )
    tm.assert_frame_equal(result, expected)


def test_merge_with_common_subexpression():
    df = DataFrame(np.random.rand(5, 2), columns=list('ab'))
    t = symbol('t', discover(df))
    expr = merge((t.a - t.a % 3).label('a'), (t.a % 3).label('b'))
    result = compute(expr, {t: df})
    expected = pd.concat(
        [
            pd.Series(df.a - df.a % 3, name='a'),
            pd.Series(df.a % 3, name='b')
        ],
        axis=1
    )
    tm.assert_frame_equal(result, expected)


def test_like():
    expr = t[t.name.like('Alice*')]
    expected = DataFrame([['Alice', 100, 1],
                          ['Alice', 50, 3]],
                         columns=['name', 'amount', 'id'])

    result = compute(expr, df).reset_index(drop=True)
    tm.assert_frame_equal(result, expected)


def test_str_len():
    expr = t.name.str.len()
    expected = pd.Series([5, 3, 5], name='name')
    result = compute(expr, df).reset_index(drop=True)
    assert_series_equal(expected, result)


def test_str_upper():
    expr = t.name.str.upper()
    expected = pd.Series(['ALICE', 'BOB', 'ALICE'], name='name')
    result = compute(expr, df).reset_index(drop=True)
    assert_series_equal(expected, result)


def test_str_lower():
    expr = t.name.str.lower()
    expected = pd.Series(['alice', 'bob', 'alice'], name='name')
    result = compute(expr, df).reset_index(drop=True)
    assert_series_equal(expected, result)

# Data-driven parametrized inputs for string-column vectorized operations.
inputs = [ # dshape, op, args, data, expected
          ('string', 'replace', ['o', 'x'], ['Someone'], ['Sxmexne']),
          ('string', 'replace', ['o', 'x', 1], ['Someone'], ['Sxmeone']),
          ('string', 'capitalize', [], ['someone'], ['Someone']),
          ('string', 'strip', [], [' someone '], ['someone']),
          ('string', 'lstrip', [], [' someone '], ['someone ']),
          ('string', 'rstrip', [], [' someone '], [' someone']),
          ('string', 'pad', [10], ['someone'], ['   someone']),
          ('string', 'pad', [10, 'left'], ['someone'], ['   someone']),
          ('string', 'pad', [10, 'right'], ['someone'], ['someone   ']),
          ('string', 'pad', [10, None, 'x'], ['someone'], ['xxxsomeone']),
          ('string', 'slice_replace', [1, 3, 'x'], ['someone'], ['sxeone']),
          ('string', 'slice_replace', [None, 3, 'x'], ['someone'], ['xeone']),
          ('string', 'slice_replace', [1, None, 'x'], ['someone'], ['sx']),
          ('string', 'slice_replace', [None, None, 'x'], ['someone'], ['x']),
          ('string', 'slice_replace', [1, None, None], ['someone'], ['s']),
          ]

# Generate inputs with missing values based on above `inputs` sequence.
na_inputs = [(datashape.Option(ds),
              op,
              args,
              data + [None],
              expected + [None]) for
             (ds, op, args, data, expected) in inputs]
# Address listcomp variable leakage in Py2.
if sys.version_info[0] == 2:
    del ds, op, args, data, expected

@pytest.mark.parametrize('ds, op, args, data, expected', inputs + na_inputs)
def test_str_ops(ds, op, args, data, expected):
    df = pd.Series(data, name='name')
    sym = symbol('t', datashape.var * datashape.R['name': ds])
    expr = getattr(sym.name.str, op)(*args)
    expected = pd.Series(expected, name='name')
    result = compute(expr, df).reset_index(drop=True)
    assert_series_equal(expected, result)
    # Test that the option / non-option dshape of the column passes through to
    # the expression's dshape.
    assert sym.dshape.measure.dict['name'].measure == expr.dshape.measure


@pytest.mark.parametrize('what,expected',
                         [('alnum', True),
                          ('alpha', True),
                          ('decimal', False),
                          ('digit', False),
                          ('lower', False),
                          ('numeric', False),
                          ('space', False),
                          ('title', True),
                          ('upper', False)])
def test_str_predicates(what, expected):
    predicate = 'is' + what
    expr = getattr(nt.name.str, predicate)()
    expected = pd.Series([expected, expected, None], name='name')
    result = compute(expr, ndf).reset_index(drop=True)
    assert_series_equal(expected, result)
    # 'discover' reports an incorrect value here...
    #assert discover(result).measure == expr.dshape.measure
    # ...so use a hardcoded one instead.
    assert str(expr.dshape.measure) == '?bool'


@pytest.mark.parametrize('slc', [0, -1, 1000,
                                 slice(0, 1), slice(None, 3),
                                 slice(-1, -2, -1),
                                 slice(-2, None)])
@pytest.mark.parametrize('df', [df, ndf])
def test_str_slice(slc, df):
    expr = t.name.str[slc]
    expected = df.name.str[slc]
    result = compute(expr, df).reset_index(drop=True)
    assert_series_equal(expected, result)


@pytest.mark.parametrize('df', [df, ndf])
def test_str_find(df):
    expr = t.name.str.find('ce')
    expected = df.name.str.find('ce')
    result = compute(expr, df).reset_index(drop=True)
    assert_series_equal(expected, result)


def test_str_cat():
    res = compute(tbig.name.str.cat(tbig.sex), dfbig)
    assert all(dfbig.name.str.cat(dfbig.sex) == res)


def test_str_cat_sep():
    res = compute(tbig.name.str.cat(tbig.sex, sep=' -- '), dfbig)
    assert all(dfbig.name.str.cat(dfbig.sex, sep=' -- ') == res)


def test_str_cat_null_row(df_add_null):
    res = compute(tbig.name.str.cat(tbig.sex, sep=' -- '), df_add_null)
    exp_res = df_add_null.name.str.cat(df_add_null.sex, sep=' -- ')

    assert all(exp_res.isnull() == res.isnull())
    assert all(exp_res[~exp_res.isnull()] == res[~res.isnull()])


def test_str_cat_chain_operation():
    expr = tbgr.name.str.cat(tbgr.comment.str.cat(tbgr.sex, sep=' --- '),
                             sep=' +++ ')
    res = compute(expr, dfbgr)
    exp_res = dfbgr.name.str.cat(dfbgr.comment.str.cat(dfbgr.sex, sep=' --- '),
                                 sep=' +++ ')
    assert all(exp_res.isnull() == res.isnull())
    assert all(exp_res[~exp_res.isnull()] == res[~res.isnull()])


def test_rowwise_by():
    f = lambda _, id, name: id + len(name)
    expr = by(t.map(f, 'int'), total=t.amount.sum())

    df = pd.DataFrame({'id': [1, 1, 2],
                       'name': ['alice', 'wendy', 'bob'],
                       'amount': [100, 200, 300.03]})
    expected = pd.DataFrame([(5, 300.03), (6, 300)], columns=expr.fields)

    result = compute(expr, df)
    tm.assert_frame_equal(result, expected)


def test_datetime_access():
    df = DataFrame({'name': ['Alice', 'Bob', 'Joe'],
                    'when': [datetime(2010, 1, 1, 1, 1, 1)] * 3,
                    'amount': [100, 200, 300],
                    'id': [1, 2, 3]})

    t = symbol('t', discover(df))

    for attr in ['day', 'month', 'minute', 'second']:
        expr = getattr(t.when, attr)
        assert_series_equal(compute(expr, df),
                            Series([1, 1, 1], name=expr._name))


@pytest.mark.parametrize('attr, args, expected',
                         [('date', [], date(2016, 6, 5)),
                          ('year', [], 2016),
                          ('month', [], 6),
                          ('day', [], 5),
                          ('time', [], time(13, 32, 1)),
                          ('hour', [], 13),
                          ('minute', [], 32),
                          ('second', [], 1),
                          ('millisecond', [], 0),
                          ('microsecond', [], 0),
                          ('nanosecond', [], 0),
                          ('week', [], 22),
                          ('weekday', [], 6),
                          ('weekday_name', [], 'Sunday'),
                          ('daysinmonth', [], 30),
                          ('weekofyear', [], 22),
                          ('dayofyear', [], 157),
                          ('dayofweek', [], 6),
                          ('quarter', [], 2),
                          ('is_month_start', [], False),
                          ('is_month_end', [], False),
                          ('is_quarter_start', [], False),
                          ('is_quarter_end', [], False),
                          ('is_year_start', [], False),
                          ('is_year_end', [], False),
                          ('days_in_month', [], 30),
                          ('strftime', ['%Y-%m-%d'], '2016-06-05'),
                         ])
def test_dt_namespace(attr, args, expected):
    df = DataFrame({'when': [datetime(2016, 6, 5, 13, 32, 1)]})
    t = symbol('t', 'var * {when: datetime}')
    expr = getattr(t.when.dt, attr)(*args)
    assert_series_equal(compute(expr, df), Series(expected, name=expr._name))


@pytest.mark.parametrize('attr, args, expected',
                         [('days', [], 7),
                          ('nanoseconds', [], 0),
                          ('seconds', [], 0),
                          ('total_seconds', [], 604800.),
                         ])
def test_td_namespace(attr, args, expected):
    """timedelta functions"""
    df = DataFrame({'span': [timedelta(7)]})
    t = symbol('t', 'var * {span: timedelta}')
    expr = getattr(t.span.dt, attr)(*args)
    assert_series_equal(compute(expr, df), Series(expected, name=expr._name))


@pytest.mark.parametrize('op, freq, expected',
                         [('round', 's', datetime(2016, 6, 5, 13, 32, 1)),
                          ('round', 'h', datetime(2016, 6, 5, 14, 0, 0)),
                          ('floor', 'h', datetime(2016, 6, 5, 13, 0, 0)),
                          ('ceil', 'h', datetime(2016, 6, 5, 14, 0, 0))])
def test_dt_round(op, freq, expected):
    df = DataFrame({'when': [datetime(2016, 6, 5, 13, 32, 1)]})
    t = symbol('t', 'var * {when: datetime}')
    expr = getattr(t.when.dt, op)(freq)
    assert_series_equal(compute(expr, df), Series(expected, name=expr._name))


def test_frame_slice():
    assert_series_equal(compute(t[0], df), df.iloc[0])
    assert_series_equal(compute(t[2], df), df.iloc[2])
    tm.assert_frame_equal(compute(t[:2], df), df.iloc[:2])
    tm.assert_frame_equal(compute(t[1:3], df), df.iloc[1:3])
    tm.assert_frame_equal(compute(t[1::2], df), df.iloc[1::2])
    tm.assert_frame_equal(compute(t[[2, 0]], df), df.iloc[[2, 0]])


def test_series_slice():
    assert compute(t.amount[0], df) == df.amount.iloc[0]
    assert compute(t.amount[2], df) == df.amount.iloc[2]
    assert_series_equal(compute(t.amount[:2], df), df.amount.iloc[:2])
    assert_series_equal(compute(t.amount[1:3], df), df.amount.iloc[1:3])
    assert_series_equal(compute(t.amount[1::2], df), df.amount.iloc[1::2])


def test_nelements():
    assert compute(t.nelements(), df) == len(df)
    assert compute(t.nrows, df) == len(df)


def test_datetime_truncation_minutes():
    data = Series(['2000-01-01T12:10:00Z', '2000-06-25T12:35:12Z'],
                  dtype='M8[ns]')
    s = symbol('s', 'var * datetime')
    result = compute(s.truncate(20, 'minutes'), data)
    expected = Series(['2000-01-01T12:00:00Z', '2000-06-25T12:20:00Z'],
                      dtype='M8[ns]', name='s')
    assert_series_equal(result, expected)


def test_datetime_truncation_nanoseconds():
    data = Series(['2000-01-01T12:10:00.000000005',
                   '2000-01-01T12:10:00.000000025'],
                  dtype='M8[ns]')
    s = symbol('s', 'var * datetime')
    expected = Series(['2000-01-01T12:10:00.000000000',
                       '2000-01-01T12:10:00.000000020'],
                      dtype='M8[ns]', name='s')
    result = compute(s.truncate(nanoseconds=20), data)
    assert_series_equal(result, expected)


def test_datetime_truncation_weeks():
    data = Series(['2000-01-01T12:10:00Z', '2000-06-25T12:35:12Z'],
                  dtype='M8[ns]')
    s = symbol('s', 'var * datetime')
    result = compute(s.truncate(2, 'weeks'), data)
    expected = Series(['1999-12-19', '2000-06-18'], dtype='M8[ns]', name='s')
    assert_series_equal(result, expected)


def test_datetime_truncation_days():
    data = Series(['2000-01-01T12:10:00Z', '2000-06-25T12:35:12Z'],
                  dtype='M8[ns]')
    s = symbol('s', 'var * datetime')
    result = compute(s.truncate(days=3), data)
    expected = Series(['1999-12-31', '2000-06-25'], dtype='M8[ns]', name='s')
    assert_series_equal(result, expected)


def test_datetime_truncation_same_as_python():
    data = Series(['2000-01-01T12:10:00Z', '2000-06-25T12:35:12Z'],
                  dtype='M8[ns]')
    s = symbol('s', 'var * datetime')
    assert (compute(s.truncate(weeks=2), data[0].to_pydatetime()) ==
            datetime(1999, 12, 26).date())


def test_complex_group_by():
    expr = by(merge(tbig.amount // 10, tbig.id % 2),
              count=tbig.name.count())
    result = compute(expr, dfbig)  # can we do this? yes we can!
    expected = dfbig.groupby([dfbig.amount // 10,
                              dfbig.id % 2])['name'].count().reset_index()
    expected = expected.rename(columns={'name': 'count'})
    tm.assert_frame_equal(result, expected)


def test_by_with_complex_summary():
    expr = by(t.name, total=t.amount.sum() + t.id.sum() - 1, a=t.id.min())
    result = compute(expr, df)
    assert list(result.columns) == expr.fields
    assert list(result.total) == [150 + 4 - 1, 200 + 2 - 1]


def test_notnull():
    assert (compute(nt.name.notnull(), ndf) == ndf.name.notnull()).all()


def test_isnan():
    assert (compute(nt.amount.isnan(), ndf) == ndf.amount.isnull()).all()


@pytest.mark.parametrize('keys', [[1], [2, 3]])
def test_isin(keys):
    expr = t[t.id.isin(keys)]
    result = compute(expr, df)
    expected = df.loc[df.id.isin(keys)]
    tm.assert_frame_equal(result, expected)


def test_nunique_table():
    expr = t.nunique()
    result = compute(expr, df)
    assert result == len(df.drop_duplicates())


def test_str_concat():
    a = Series(('a', 'b', 'c'))
    s = symbol('s', "3 * string[1, 'U32']")
    expr = s + 'a'
    assert (compute(expr, a) == (a + 'a')).all()


def test_str_repeat():
    a = Series(('a', 'b', 'c'))
    s = symbol('s', "3 * string[1, 'U32']")
    expr = s.repeat(3)
    assert (compute(expr, a) == (a * 3)).all()


def test_str_interp():
    a = Series(('%s', '%s', '%s'))
    s = symbol('s', "3 * string[1, 'U32']")
    expr = s.interp(1)
    assert (compute(expr, a) == (a % 1)).all()


def test_timedelta_arith():
    series = Series(pd.date_range('2014-01-01', '2014-02-01'))
    sym = symbol('s', discover(series))
    delta = timedelta(days=1)
    assert (compute(sym + delta, series) == series + delta).all()
    assert (compute(sym - delta, series) == series - delta).all()
    assert (
        compute(sym - (sym - delta), series) ==
        series - (series - delta)
    ).all()


@pytest.mark.parametrize('func,expected', (
    ('var', timedelta(0, 8, 250000)),
    ('std', timedelta(0, 2, 872281)),
))
def test_timedelta_stat_reduction(func, expected):
    deltas = pd.Series([timedelta(seconds=n) for n in range(10)])
    sym = symbol('s', discover(deltas))
    assert compute(getattr(sym, func)(), deltas) == expected


def test_coerce_series():
    s = pd.Series(list('123'), name='a')
    t = symbol('t', discover(s))
    result = compute(t.coerce(to='int64'), s)
    expected = pd.Series([1, 2, 3], name=s.name)
    assert_series_equal(result, expected)


@pytest.mark.parametrize('d, tp, ptp', [(list(range(10)), 'string', str),
                                        (list(range(10)) + [None], '?string', str),
                                        (['2010-01-01'], 'datetime', np.datetime64),
                                        (['2010-01-01', None, ''], '?datetime', np.datetime64)])
def test_coerce_series_string_datetime(d, tp, ptp):
    s = pd.Series(d, name='a')
    e = symbol('t', discover(s)).coerce(to=tp)
    assert e.schema == dshape(tp)
    result = compute(e, s)
    expected = s.astype(ptp)
    assert_series_equal(result, expected)


def test_concat_arr():
    s_data = Series(np.arange(15))
    t_data = Series(np.arange(15, 30))

    s = symbol('s', discover(s_data))
    t = symbol('t', discover(t_data))

    assert (
        compute(concat(s, t), {s: s_data, t: t_data}) == Series(np.arange(30))
    ).all()


def test_concat_mat():
    s_data = DataFrame(np.arange(15).reshape(5, 3), columns=list('abc'))
    t_data = DataFrame(np.arange(15, 30).reshape(5, 3), columns=list('abc'))

    s = symbol('s', discover(s_data))
    t = symbol('t', discover(t_data))

    tm.assert_frame_equal(
        compute(concat(s, t), {s: s_data, t: t_data}),
        pd.DataFrame(np.arange(30).reshape(10, 3), columns=list('abc')),
    )


def test_count_keepdims_frame():
    df = pd.DataFrame(dict(a=[1, 2, 3, np.nan]))
    s = symbol('s', discover(df))
    assert_series_equal(compute(s.count(keepdims=True), df),
                        pd.Series([df.shape[0]], name='s_count'))


def test_time_field():
    data = pd.Series(pd.date_range(start='20120101', end='20120102', freq='H'))
    s = symbol('s', discover(data))
    result = compute(s.time, data)
    expected = data.dt.time
    expected.name = 's_time'
    assert_series_equal(result, expected)



@pytest.mark.parametrize('n', [-1, 0, 1])
def test_shift(n):
    data = pd.Series(pd.date_range(start='20120101', end='20120102', freq='H'))
    s = symbol('s', discover(data))
    result = compute(s.shift(n), data)
    expected = data.shift(n)
    assert_series_equal(result, expected)


def test_selection_inner_inputs():
    s_data = pd.DataFrame({'a': np.arange(5)})
    t_data = pd.DataFrame({'a': np.arange(5)})

    s = symbol('s', 'var * {a: int64}')
    t = symbol('t', 'var * {a: int64}')

    tm.assert_frame_equal(
        compute(s[s.a == t.a], {s: s_data, t: t_data}),
        s_data
    )


def test_by_with_reduction_on_df():
    expr = by(tbig.name, id_sum=tbig.id.sum(), count=tbig.count())
    compute(expr, dfbig)


def test_coalesce():
    data = pd.Series([0, None, 1, None, 2, None], dtype=object)

    s = symbol('s', 'var * ?int')
    t = symbol('t', 'int')
    u = symbol('u', '?int')
    v = symbol('v', 'var * int')
    w = symbol('w', 'var * ?int')

    # array to scalar
    tm.assert_series_equal(
        compute(coalesce(s, t), {s: data, t: -1}),
        pd.Series([0, -1, 1, -1, 2, -1], dtype=object),
    )
    # array to scalar with NULL
    tm.assert_series_equal(
        compute(coalesce(s, u), {s: data, u: None}),
        pd.Series([0, None, 1, None, 2, None], dtype=object),
    )
    # array to array
    tm.assert_series_equal(
        compute(coalesce(s, v), {
            s: data, v: np.array([-1, -2, -3, -4, -5, -6]),
        }),
        pd.Series([0, -2, 1, -4, 2, -6], dtype=object),
    )
    # array to array with NULL
    tm.assert_series_equal(
        compute(coalesce(s, w), {
            s: data, w: np.array([-1, None, -3, -4, -5, -6]),
        }),
        pd.Series([0, None, 1, -4, 2, -6], dtype=object),
    )



from __future__ import absolute_import, division, print_function

import fnmatch

from toolz import compose, identity
from datashape.predicates import isscalar

from ..expr import (
    Expr, ElemWise, SimpleSelection, Sort, Apply, Distinct, Join, By, Label,
    Summary, by, ReLabel, Like, Reduction, Head
)
from .python import (
    compute, rrowfunc, rowfunc, pair_assemble, reduce_by_funcs, binops
)
from ..expr.broadcast import broadcast_collect
from ..expr.optimize import simple_selections
from ..compatibility import builtins, unicode
from ..expr import reductions
from ..dispatch import dispatch

from .core import compute_up

import py4j
from pyspark import SparkContext
import pyspark
from pyspark.rdd import RDD

__all__ = ['RDD', 'pyspark', 'SparkContext']


# PySpark adds a SIGCHLD signal handler, but that breaks other packages, so we
# remove it
# See https://issues.apache.org/jira/browse/SPARK-1394
try:
    import signal
    signal.signal(signal.SIGCHLD, signal.SIG_DFL)
except:
    pass


@dispatch(Expr, RDD)
def optimize(expr, seq):
    return simple_selections(broadcast_collect(expr))


@dispatch(ElemWise, RDD)
def compute_up(t, rdd, **kwargs):
    func = rowfunc(t)
    return rdd.map(func)


@dispatch(SimpleSelection, RDD)
def compute_up(t, rdd, **kwargs):
    predicate = optimize(t.predicate, rdd)
    predicate = rrowfunc(predicate, t._child)
    return rdd.filter(predicate)


rdd_reductions = {
    reductions.sum: RDD.sum,
    reductions.min: RDD.min,
    reductions.max: RDD.max,
    reductions.count: RDD.count,
    reductions.mean: RDD.mean,
    reductions.var: RDD.variance,
    reductions.std: RDD.stdev,
    reductions.nunique: compose(RDD.count, RDD.distinct)
}


@dispatch(tuple(rdd_reductions), RDD)
def compute_up(t, rdd, **kwargs):
    return rdd_reductions[type(t)](rdd)


def istruthy(x):
    return not not x


@dispatch(reductions.any, RDD)
def compute_up(t, rdd, **kwargs):
    return istruthy(rdd.filter(identity).take(1))


@dispatch(reductions.all, RDD)
def compute_up(t, rdd, **kwargs):
    return not rdd.filter(lambda x: not x).take(1)


@dispatch(Head, RDD)
def compute_up(t, rdd, **kwargs):
    return rdd.take(t.n)


@dispatch(Apply, RDD)
def compute_up(t, rdd, **kwargs):
    if t._splittable:
        return rdd.mapPartitions(t.func)
    else:
        raise NotImplementedError("Can only apply splittable functions."
                                  "To apply function to each partition add "
                                  "splittable=True kwarg to call to apply.  "
                                  "t.apply(func, dshape, splittable=True)")


@dispatch(Sort, RDD)
def compute_up(t, rdd, **kwargs):
    if isinstance(t.key, (str, unicode, tuple, list)):
        key = rowfunc(t._child[t.key])
    else:
        key = optimize(t.key, rdd)
        key = rrowfunc(key, t._child)
    return (rdd.keyBy(key)
            .sortByKey(ascending=t.ascending)
            .map(lambda x: x[1]))


@dispatch(Distinct, RDD)
def compute_up(t, rdd, **kwargs):
    if t.on:
        raise NotImplementedError(
            'spark backend cannot specify what columns to distinct on'
        )
    return rdd.distinct()


def jgetattr(data, attr, default=None):
    """Spark's API doesn't properly implement the ``getattr`` interface, so
    we work around it.
    """
    try:
        return getattr(data, attr, default)
    except py4j.protocol.Py4JJavaError:
        return default


@compute_up.register(Join, RDD, RDD)
def spark_join(t, lhs, rhs, **kwargs):
    on_left = rowfunc(t.lhs[t.on_left])
    on_right = rowfunc(t.rhs[t.on_right])

    lhs = lhs.keyBy(on_left)
    rhs = rhs.keyBy(on_right)

    how = t.how

    if how == 'inner':
        joiner = lhs.join
    elif how == 'left':
        joiner = lhs.leftOuterJoin
    elif how == 'right':
        joiner = lhs.rightOuterJoin
    elif how == 'outer':
        joiner = lhs.fullOuterJoin
    else:
        raise ValueError("Invalid join type %r, must be one of "
                         "{'inner', 'left', 'right', 'outer'}" % how)

    rdd = joiner(rhs)
    assemble = pair_assemble(t)

    return rdd.map(lambda x: assemble(x[1]))


@dispatch(By, RDD)
def compute_up(t, rdd, **kwargs):
    grouper = optimize(t.grouper, rdd)
    apply = optimize(t.apply, rdd)
    t = by(grouper, apply)
    if ((isinstance(t.apply, Reduction) and type(t.apply) in binops) or
        (isinstance(t.apply, Summary) and
         builtins.all(type(val) in binops for val in t.apply.values))):
        grouper, binop, combiner, initial = reduce_by_funcs(t)

        if isscalar(t.grouper.dshape.measure):
            keyfunc = lambda x: (x,)
        else:
            keyfunc = identity
        if isscalar(t.apply.dshape.measure):
            valfunc = lambda x: (x,)
        else:
            valfunc = identity
        unpack = lambda kv: keyfunc(kv[0]) + valfunc(kv[1])

        create = lambda v: binop(initial, v)

        return (rdd.keyBy(grouper)
                   .combineByKey(create, binop, combiner)
                   .map(unpack))
    else:
        raise NotImplementedError("By only implemented for common reductions."
                                  "\nGot %s" % type(t.apply))


@dispatch((Label, ReLabel), RDD)
def compute_up(t, rdd, **kwargs):
    return rdd


@dispatch(Summary, RDD)
def compute_up(t, rdd, **kwargs):
    rdd = rdd.cache()
    return tuple(
        compute(value, {t._child: rdd}, return_type='native')
        for value in t.values
    )


@dispatch(Like, RDD)
def compute_up(t, rdd, **kwargs):
    def func(value, pattern=t.pattern):
        return fnmatch.fnmatch(value, pattern)
    return rdd.map(func)



from __future__ import absolute_import, division, print_function

from collections import Callable

from blaze.dispatch import dispatch
from blaze.compatibility import u8
from functools import partial

from .json_dumps import json_dumps


json_dumps_trusted_ns = dict()
dispatch = partial(dispatch, namespace=json_dumps_trusted_ns)


@dispatch(Callable)
def json_dumps_trusted(f):
    # let the server serialize any callable - this is only used for testing
    # at present - do the error handling when json comes from client so in
    # object_hook, catch anything that is not pandas_numpy
    fcn = u8(".".join([f.__module__, f.__name__]))
    return {u'__!callable': fcn}


for tps, func in json_dumps.funcs.items():
    json_dumps_trusted.register(tps)(func)



from __future__ import absolute_import, division, print_function

from collections import Mapping, OrderedDict
import datetime
from functools import reduce, partial
import inspect
from itertools import repeat
import numbers

from pprint import pformat
from weakref import WeakValueDictionary

import toolz
from toolz import unique, concat, first

from ..compatibility import _strtypes
from ..dispatch import dispatch
from ..utils import ordered_intersect

__all__ = ['Node', 'path', 'common_subexpression', 'eval_str']


base = (numbers.Number,) + _strtypes + (datetime.datetime, datetime.timedelta)


def _resolve_args(cls, *args, **kwargs):
    """Resolve the arguments from a node class into an ordereddict.
    All arguments are assumed to have a default of None.

    This is sort of like getargspec but uses the `Node` specific machinery.

    Parameters
    ----------
    cls : subclass of Node
        The class to resolve the arguments for.
    *args, **kwargs
        The arguments that were passed.

    Returns
    -------
    args : OrderedDict
        A dictionary mapping argument names to their value in the order
        they appear in the `_arguments` tuple.

    Examples
    --------
    >>> class MyNode(Node):
    ...     _arguments = 'a', 'b', 'c'
    ...

    good cases
    >>> _resolve_args(MyNode, 1, 2, 3)
    OrderedDict([('a', 1), ('b', 2), ('c', 3)])
    >>> _resolve_args(MyNode, 1, 2, c=3)
    OrderedDict([('a', 1), ('b', 2), ('c', 3)])
    >>> _resolve_args(MyNode, a=1, b=2, c=3)
    OrderedDict([('a', 1), ('b', 2), ('c', 3)])

    error cases
    >>> _resolve_args(MyNode, 1, 2, 3, a=4)
    Traceback (most recent call last):
       ...
    TypeError: MyNode got multiple values for argument 'a'
    >>> _resolve_args(MyNode, 1, 2, 3, 4)
    Traceback (most recent call last):
       ...
    TypeError: MyNode takes 3 positional arguments but 4 were given
    >>> _resolve_args(MyNode, 1, 2, 3, d=4)
    Traceback (most recent call last):
       ...
    TypeError: MyNode got unknown keywords: d
    """
    attrs = cls._arguments
    attrset = set(attrs)
    if not set(kwargs) <= attrset:
        raise TypeError(
            '%s got unknown keywords: %s' % (
                cls.__name__,
                ', '.join(set(kwargs) - attrset),
            ),
        )

    if len(args) > len(attrs):
        raise TypeError(
            '%s takes 3 positional arguments but %d were given' % (
                cls.__name__,
                len(args),
            ),
        )

    attributes = OrderedDict(zip(attrs, repeat(None)))
    to_add = dict(zip(attrs, args))
    attributes.update(to_add)
    added = set(to_add)

    for key, value in kwargs.items():
        if key in added:
            raise TypeError(
                '%s got multiple values for argument %r' % (
                    cls.__name__,
                    key,
                ),
            )
        attributes[key] = value
        added.add(key)

    return attributes


def _static_identity(ob):
    return type(ob)._static_identity(*ob._args)


def _setattr(ob, name, value):
    object.__setattr__(ob, name, value)
    return value


class Node(object):
    """ Node in a tree

    This serves as the base class for ``Expr``.  This class holds all of the
    tree traversal functions that are independent of tabular or array
    computation.  This is everything that we can do independent of the problem
    domain.  Note that datashape is not imported.

    See Also
    --------

    blaze.expr.expressions.Expr
    """
    _arguments = '_child',
    __inputs__ = '_child',
    __expr_instance_cache = WeakValueDictionary()

    def __new__(cls, *args, **kwargs):
        static_id = cls._static_identity(*args, **kwargs)
        try:
            return cls.__expr_instance_cache[static_id]
        except KeyError:
            cls.__expr_instance_cache[static_id] = self = super(
                Node,
                cls,
            ).__new__(cls)._init(*args, **kwargs)
            return self

    def _init(self, *args, **kwargs):
        for name, arg in _resolve_args(type(self), *args, **kwargs).items():
            _setattr(self, name, arg)

        _setattr(self, '_hash', None)
        return self

    def __setattr__(self, name, value):
        raise AttributeError('cannot set attributes of immutable objects')

    @property
    def _args(self):
        return tuple(getattr(self, slot) for slot in self._arguments)

    @classmethod
    def _static_identity(cls, *args, **kwargs):
        return (cls,) + tuple(_resolve_args(cls, *args, **kwargs).values())

    @property
    def _inputs(self):
        return tuple(getattr(self, i) for i in self.__inputs__)

    def _leaves(self):
        """ Leaves of an expression tree

        All nodes without inputs.  Leaves are returned in order, left to right.

        >>> from blaze.expr import symbol, join, by

        >>> t = symbol('t', 'var * {id: int32, name: string}')
        >>> t._leaves()
        [<`t` symbol; dshape='var * {id: int32, name: string}'>]
        >>> by(t.name, count=t.id.nunique())._leaves()
        [<`t` symbol; dshape='var * {id: int32, name: string}'>]

        >>> v = symbol('v', 'var * {id: int32, city: string}')
        >>> join(t, v)._leaves() == [t, v]
        True
        """

        if not self._inputs:
            return [self]
        else:
            return list(unique(concat(i._leaves() for i in self._inputs if
                                      isinstance(i, Node))))

    def isidentical(self, other):
        """Identity check for blaze expressions.
        """
        return self is other

    def __hash__(self):
        hash_ = self._hash
        if hash_ is None:
            hash_ = _setattr(
                self,
                '_hash',
                hash((type(self), _static_identity(self))),
            )
        return hash_

    def __str__(self):
        rep = [
            '%s=%s' % (slot, _str(arg))
            for slot, arg in zip(self._arguments, self._args)
        ]
        return '%s(%s)' % (type(self).__name__, ', '.join(rep))

    def _traverse(self):
        """ Traverse over tree, yielding all subtrees and leaves """
        yield self
        traversals = (
            arg._traverse() if isinstance(arg, Node) else [arg]
            for arg in self._args
        )
        for item in concat(traversals):
            yield item

    def _subs(self, d):
        """ Substitute terms in the tree

        >>> from blaze.expr import symbol
        >>> t = symbol('t', 'var * {name: string, amount: int, id: int}')
        >>> expr = t.amount + 3
        >>> expr._subs({3: 4, 'amount': 'id'}).isidentical(t.id + 4)
        True
        """
        return subs(self, d)

    def _resources(self):
        return toolz.merge([arg._resources() for arg in self._args
                            if isinstance(arg, Node)])

    def _subterms(self):
        return subterms(self)

    def __contains__(self, other):
        return other in set(self._subterms())

    def __reduce_ex__(self, protocol):
        if protocol < 2:
            raise ValueError(
                'blaze expressions may only be pickled with protocol'
                ' 2 or greater',
            )
        return type(self), self._args

    def __eq__(self, other):
        try:
            return self.isidentical(other) or self._eq(other)
        except AttributeError:
            return False

    def __ne__(self, other):
        return self._ne(other)

    def __lt__(self, other):
        return self._lt(other)

    def __le__(self, other):
        return self._le(other)

    def __gt__(self, other):
        return self._gt(other)

    def __ge__(self, other):
        return self._ge(other)

    def __add__(self, other):
        return self._add(other)

    def __radd__(self, other):
        return self._radd(other)

    def __mul__(self, other):
        return self._mul(other)

    def __rmul__(self, other):
        return self._rmul(other)

    def __div__(self, other):
        return self._div(other)

    def __rdiv__(self, other):
        return self._rdiv(other)

    __truediv__ = __div__
    __rtruediv__ = __rdiv__

    def __floordiv__(self, other):
        return self._floordiv(other)

    def __rfloordiv__(self, other):
        return self._rfloordiv(other)

    def __sub__(self, other):
        return self._sub(other)

    def __rsub__(self, other):
        return self._rsub(other)

    def __pow__(self, other):
        return self._pow(other)

    def __rpow__(self, other):
        return self._rpow(other)

    def __mod__(self, other):
        return self._mod(other)

    def __rmod__(self, other):
        return self._rmod(other)

    def __or__(self, other):
        return self._or(other)

    def __ror__(self, other):
        return self._ror(other)

    def __and__(self, other):
        return self._and(other)

    def __rand__(self, other):
        return self._rand(other)

    def __neg__(self):
        return self._neg()

    def __invert__(self):
        return self._invert()

    def __abs__(self):
        from .math import abs
        return abs(self)


def get_callable_name(o):
    """Welcome to str inception. Leave your kittens at home.
    """
    # special case partial objects
    if isinstance(o, partial):
        keywords = o.keywords
        kwds = (
            ', '.join('%s=%r' % item for item in keywords.items())
            if keywords else
            ''
        )
        args = ', '.join(map(repr, o.args))
        arguments = []
        if args:
            arguments.append(args)
        if kwds:
            arguments.append(kwds)
        return 'partial(%s, %s)' % (
            get_callable_name(o.func),
            ', '.join(arguments),
        )

    try:
        # python 3 makes builtins look nice
        return o.__qualname__
    except AttributeError:
        try:
            # show the module of the object, if we can
            return '%s.%s' % (inspect.getmodule(o).__name__, o.__name__)
        except AttributeError:
            try:
                # __self__ tells us the class the method is bound to
                return '%s.%s' % (o.__self__.__name__, o.__name__)
            except AttributeError:
                # exhausted all avenues of printing callables so just print the
                # name of the object
                return o.__name__


def _str(s):
    """ Wrap single quotes around strings """
    if isinstance(s, str):
        return repr(s)
    elif callable(s):
        return get_callable_name(s)
    elif isinstance(s, Node):
        return str(s)
    elif isinstance(s, (list, tuple)):
        body = ", ".join(_str(x) for x in s)
        return "({0})".format(body if len(s) > 1 else (body + ","))
    else:
        return pformat(s).rstrip()


@dispatch(Node)
def subterms(expr):
    return concat([[expr], concat(map(subterms, expr._inputs))])


@dispatch(object)
def subterms(x):
    yield x


def subs(o, d):
    """ Substitute values within data structure

    >>> subs(1, {1: 2})
    2

    >>> subs([1, 2, 3], {2: 'Hello'})
    [1, 'Hello', 3]
    """
    d = {k: v for k, v in d.items() if k is not v}
    if not d:
        return o
    try:
        if o in d:
            d = d.copy()
            o = d.pop(o)
    except TypeError:
        pass
    return _subs(o, d)


@dispatch((tuple, list), Mapping)
def _subs(o, d):
    return type(o)(subs(arg, d) for arg in o)


@dispatch(Node, Mapping)
def _subs(o, d):
    """

    >>> from blaze.expr import symbol
    >>> t = symbol('t', 'var * {name: string, balance: int}')
    >>> subs(t, {'balance': 'amount'}).fields
    ['name', 'amount']
    """
    newargs = (subs(arg, d) for arg in o._args)
    return type(o)(*newargs)


@dispatch(object, Mapping)
def _subs(o, d):
    """ Private dispatched version of ``subs``

    >>> subs('Hello', {})
    'Hello'
    """
    return o


def path(a, b):
    """ A path of nodes from a to b

    >>> from blaze.expr import symbol
    >>> t = symbol('t', 'var * {name: string, amount: int, id: int}')
    >>> expr = t.amount.sum()
    >>> list(path(expr, t))
    [sum(t.amount), t.amount, <`t` symbol; dshape='...'>]
    """
    while not a.isidentical(b):
        yield a
        if not a._inputs:
            break
        for child in a._inputs:
            if any(b.isidentical(node) for node in child._traverse()):
                a = child
                break
    yield a


def common_subexpression(expr, *exprs):
    """ Common sub expression between subexpressions

    Examples
    --------

    >>> from blaze.expr import symbol
    >>> t = symbol('t', 'var * {x: int, y: int}')
    >>> common_subexpression(t.x, t.y)
    <`t` symbol; dshape='var * {x: int32, y: int32}'>
    """
    # only one expression has itself as a common subexpression
    if not exprs:
        return expr

    exprs = (expr,) + exprs

    # get leaves for every expression
    all_leaves = [expr._leaves() for expr in exprs]

    # leaves common to all expressions
    leaves = set.intersection(*map(set, all_leaves))

    # no common leaves therefore no common subexpression
    if not leaves:
        raise ValueError(
            'No common leaves found in expressions %s' % list(exprs)
        )

    # list of paths from each expr to each leaf
    pathlist = [list(path(expr, leaf)) for expr in exprs for leaf in leaves]

    # ordered intersection of paths
    common = reduce(ordered_intersect, pathlist)
    if not common:
        raise ValueError(
            'No common subexpression found in paths to leaf: %s' % list(
                map(set, pathlist)
            )
        )

    # the first expression is the deepest node in the tree that is an ancestor
    # of every expression in `exprs`
    return first(common)


def eval_str(expr):
    """ String suitable for evaluation

    >>> from blaze.expr import symbol, eval_str
    >>> x = symbol('x', 'real')
    >>> eval_str(2*x + 1)
    '(2 * x) + 1'

    >>> from datetime import date
    >>> eval_str(date(2000, 1, 20))
    'datetime.date(2000, 1, 20)'
    """
    from datetime import date, datetime
    if isinstance(expr, (date, datetime)):
        return repr(expr)
    return repr(expr) if isinstance(expr, _strtypes) else str(expr)


def parenthesize(s):
    """

    >>> parenthesize('1')
    '1'
    >>> parenthesize('1 + 2')
    '(1 + 2)'
    """
    if ' ' in s:
        return '(%s)' % s
    else:
        return s
